"""Tests tidy3d/components/data/monitor_data.py"""
import numpy as np
import matplotlib.pyplot as plt
import pytest
import pydantic.v1 as pydantic
import tidy3d as td

from tidy3d.exceptions import DataError

from tidy3d.components.data.data_array import FreqModeDataArray
from tidy3d.components.data.monitor_data import FieldData, FieldTimeData, PermittivityData

from tidy3d.components.data.monitor_data import ModeSolverData, ModeData
from tidy3d.components.data.monitor_data import FluxData, FluxTimeData, DiffractionData

from .test_data_arrays import make_scalar_field_data_array, make_scalar_field_time_data_array
from .test_data_arrays import make_scalar_mode_field_data_array
from .test_data_arrays import make_scalar_mode_field_data_array_smooth
from .test_data_arrays import make_flux_data_array, make_flux_time_data_array
from .test_data_arrays import make_mode_amps_data_array, make_mode_index_data_array
from .test_data_arrays import make_diffraction_data_array
from .test_data_arrays import FIELD_MONITOR, FIELD_TIME_MONITOR, MODE_SOLVE_MONITOR
from .test_data_arrays import MODE_MONITOR, PERMITTIVITY_MONITOR, FLUX_MONITOR, FLUX_TIME_MONITOR
from .test_data_arrays import FIELD_MONITOR_2D, FIELD_TIME_MONITOR_2D
from .test_data_arrays import DIFFRACTION_MONITOR, SIM_SYM, SIM
from ..utils import assert_log_level, log_capture

# data array instances
AMPS = make_mode_amps_data_array()
N_COMPLEX = make_mode_index_data_array()
FLUX = make_flux_data_array()
FLUX_TIME = make_flux_time_data_array()
GRID_CORRECTION = FreqModeDataArray(
    1 + 0.01 * np.random.rand(*N_COMPLEX.shape), coords=N_COMPLEX.coords
)

""" Make the montor data """


def make_field_data(symmetry: bool = True):
    sim = SIM_SYM if symmetry else SIM
    return FieldData(
        monitor=FIELD_MONITOR,
        Ex=make_scalar_field_data_array("Ex", symmetry),
        Ey=make_scalar_field_data_array("Ey", symmetry),
        Ez=make_scalar_field_data_array("Ez", symmetry),
        Hx=make_scalar_field_data_array("Hx", symmetry),
        Hz=make_scalar_field_data_array("Hz", symmetry),
        symmetry=sim.symmetry,
        symmetry_center=sim.center,
        grid_expanded=sim.discretize_monitor(FIELD_MONITOR),
    )


def make_field_time_data(symmetry: bool = True):
    sim = SIM_SYM if symmetry else SIM
    return FieldTimeData(
        monitor=FIELD_TIME_MONITOR,
        Ex=make_scalar_field_time_data_array("Ex", symmetry),
        Ey=make_scalar_field_time_data_array("Ey", symmetry),
        Ez=make_scalar_field_time_data_array("Ez", symmetry),
        Hz=make_scalar_field_time_data_array("Hz", symmetry),
        Hx=make_scalar_field_time_data_array("Hx", symmetry),
        symmetry=sim.symmetry,
        symmetry_center=sim.center,
        grid_expanded=sim.discretize_monitor(FIELD_TIME_MONITOR),
    )


def make_field_data_2d(symmetry: bool = True):
    sim = SIM_SYM if symmetry else SIM
    return FieldData(
        monitor=FIELD_MONITOR_2D,
        Ex=make_scalar_field_data_array("Ex", symmetry).interp(y=[0]),
        Ey=make_scalar_field_data_array("Ey", symmetry).interp(y=[0]),
        Ez=make_scalar_field_data_array("Ez", symmetry).interp(y=[0]),
        Hx=make_scalar_field_data_array("Hx", symmetry).interp(y=[0]),
        Hz=make_scalar_field_data_array("Hz", symmetry).interp(y=[0]),
        symmetry=sim.symmetry,
        symmetry_center=sim.center,
        grid_expanded=sim.discretize_monitor(FIELD_MONITOR_2D),
    )


def make_field_time_data_2d(symmetry: bool = True):
    sim = SIM_SYM if symmetry else SIM
    return FieldTimeData(
        monitor=FIELD_TIME_MONITOR_2D,
        Ex=make_scalar_field_time_data_array("Ex", symmetry).interp(y=[0]),
        Ey=make_scalar_field_time_data_array("Ey", symmetry).interp(y=[0]),
        Ez=make_scalar_field_time_data_array("Ez", symmetry).interp(y=[0]),
        Hx=make_scalar_field_time_data_array("Hx", symmetry).interp(y=[0]),
        Hz=make_scalar_field_time_data_array("Hz", symmetry).interp(y=[0]),
        symmetry=sim.symmetry,
        symmetry_center=sim.center,
        grid_expanded=sim.discretize_monitor(FIELD_TIME_MONITOR_2D),
    )


def make_mode_solver_data():
    mode_data = ModeSolverData(
        monitor=MODE_SOLVE_MONITOR,
        Ex=make_scalar_mode_field_data_array("Ex"),
        Ey=make_scalar_mode_field_data_array("Ey"),
        Ez=make_scalar_mode_field_data_array("Ez"),
        Hx=make_scalar_mode_field_data_array("Hx"),
        Hy=make_scalar_mode_field_data_array("Hy"),
        Hz=make_scalar_mode_field_data_array("Hz"),
        symmetry=SIM_SYM.symmetry,
        symmetry_center=SIM_SYM.center,
        grid_expanded=SIM_SYM.discretize_monitor(MODE_SOLVE_MONITOR),
        n_complex=N_COMPLEX.copy(),
        grid_primal_correction=GRID_CORRECTION,
        grid_dual_correction=GRID_CORRECTION,
    )
    # Mode solver data needs to be normalized
    scaling = np.sqrt(np.abs(mode_data.symmetry_expanded_copy.flux))
    norm_data_dict = {key: val / scaling for key, val in mode_data.field_components.items()}
    mode_data_norm = mode_data.copy(update=norm_data_dict)
    return mode_data_norm


def make_mode_solver_data_smooth():
    mode_data = ModeSolverData(
        monitor=MODE_SOLVE_MONITOR,
        Ex=make_scalar_mode_field_data_array_smooth("Ex", rot=0.13 * np.pi),
        Ey=make_scalar_mode_field_data_array_smooth("Ey", rot=0.26 * np.pi),
        Ez=make_scalar_mode_field_data_array_smooth("Ez", rot=0.39 * np.pi),
        Hx=make_scalar_mode_field_data_array_smooth("Hx", rot=0.52 * np.pi),
        Hy=make_scalar_mode_field_data_array_smooth("Hy", rot=0.65 * np.pi),
        Hz=make_scalar_mode_field_data_array_smooth("Hz", rot=0.78 * np.pi),
        symmetry=SIM_SYM.symmetry,
        symmetry_center=SIM_SYM.center,
        grid_expanded=SIM_SYM.discretize_monitor(MODE_SOLVE_MONITOR),
        n_complex=N_COMPLEX.copy(),
        grid_primal_correction=GRID_CORRECTION,
        grid_dual_correction=GRID_CORRECTION,
    )
    # Mode solver data needs to be normalized
    scaling = np.sqrt(np.abs(mode_data.symmetry_expanded_copy.flux))
    norm_data_dict = {key: val / scaling for key, val in mode_data.field_components.items()}
    mode_data_norm = mode_data.copy(update=norm_data_dict)
    return mode_data_norm


def make_permittivity_data(symmetry: bool = True):
    sim = SIM_SYM if symmetry else SIM
    return PermittivityData(
        monitor=PERMITTIVITY_MONITOR,
        eps_xx=make_scalar_field_data_array("Ex", symmetry),
        eps_yy=make_scalar_field_data_array("Ey", symmetry),
        eps_zz=make_scalar_field_data_array("Ez", symmetry),
        symmetry=sim.symmetry,
        symmetry_center=sim.center,
        grid_expanded=sim.discretize_monitor(PERMITTIVITY_MONITOR),
    )


def make_mode_data():
    return ModeData(monitor=MODE_MONITOR, amps=AMPS.copy(), n_complex=N_COMPLEX.copy())


def make_flux_data():
    return FluxData(monitor=FLUX_MONITOR, flux=FLUX.copy())


def make_flux_time_data():
    return FluxTimeData(monitor=FLUX_TIME_MONITOR, flux=FLUX_TIME.copy())


def make_diffraction_data():
    sim_size, bloch_vecs, data = make_diffraction_data_array()
    return DiffractionData(
        monitor=DIFFRACTION_MONITOR,
        Etheta=data,
        Ephi=data,
        Er=data,
        Htheta=data,
        Hphi=data,
        Hr=data,
        sim_size=sim_size,
        bloch_vecs=bloch_vecs,
    )


""" Test them out """


def test_field_data():
    data = make_field_data()
    # Check that calling flux and dot on 3D data raise errors
    with pytest.raises(DataError):
        _ = data.dot(data)
    data_2d = make_field_data_2d()
    for field in FIELD_MONITOR.fields:
        _ = getattr(data_2d, field)
    # Compute flux directly
    flux1 = np.abs(data_2d.flux)
    # Compute flux as dot product with itself
    flux2 = np.abs(data_2d.dot(data_2d))
    # Assert result is the same
    assert np.all(flux1 == flux2)


def test_field_data_to_source():
    data = make_field_data_2d(symmetry=True)
    data = data.copy(update={key: val.isel(f=[-1]) for key, val in data.field_components.items()})
    _ = data.to_source(source_time=td.GaussianPulse(freq0=2e14, fwidth=2e13), center=(1, 2, 3))
    data = make_field_data_2d(symmetry=False)
    data = data.copy(update={key: val.isel(f=[-1]) for key, val in data.field_components.items()})
    _ = data.to_source(source_time=td.GaussianPulse(freq0=2e14, fwidth=2e13), center=(1, 2, 3))


def test_field_time_data():
    data = make_field_time_data_2d()
    for field in FIELD_TIME_MONITOR.fields:
        _ = getattr(data, field)
    # Check that flux can be computed
    _ = np.abs(data.flux)
    # Check that trying to call the dot product raises an error for time data
    with pytest.raises(DataError):
        _ = data.dot(data)


def test_mode_solver_data():
    data = make_mode_solver_data()
    for field in "EH":
        for component in "xyz":
            _ = getattr(data, field + component)
    # Compute flux directly
    flux1 = np.abs(data.flux)
    # Compute flux as dot product with itself
    flux2 = np.abs(data.dot(data))
    # Assert result is the same
    assert np.all(flux1 == flux2)
    # Compute dot product with a field data
    field_data = make_field_data_2d()
    dot = data.dot(field_data)
    # Check that broadcasting worked
    assert data.Ex.f == dot.f
    assert data.Ex.mode_index == dot.mode_index
    # Also try with a feild data at a single frequency that is not in the data frequencies
    freq = 0.9 * field_data.Ex.f[0]
    fields = field_data.field_components.items()
    fields_single_f = {key: val.isel(f=[0]).assign_coords(f=[freq]) for key, val in fields}
    field_data = field_data.copy(update=fields_single_f)
    dot = data.dot(field_data)
    # Check that broadcasting worked
    assert data.Ex.f == dot.f
    assert data.Ex.mode_index == dot.mode_index
    # Check eps_spec validator
    num_freqs = len(data.monitor.freqs)
    _ = data.updated_copy(eps_spec=["diagonal"] * num_freqs)
    _ = data.updated_copy(eps_spec=["tensorial_real"] * num_freqs)
    _ = data.updated_copy(eps_spec=["tensorial_complex"] * num_freqs)
    # wrong keyword
    with pytest.raises(pydantic.ValidationError):
        _ = data.updated_copy(eps_spec=["tensorial"] * num_freqs)
    # wrong number
    with pytest.raises(pydantic.ValidationError):
        _ = data.updated_copy(eps_spec=["diagonal"] * (num_freqs + 1))
    # check monitor direction changes upon time reversal
    data_reversed = data.time_reversed_copy
    assert data_reversed.monitor.direction == "-"


def test_permittivity_data():
    data = make_permittivity_data()
    for comp in "xyz":
        _ = getattr(data, "eps_" + comp + comp)


def test_mode_data():
    data = make_mode_data()
    _ = data.amps
    _ = data.n_complex
    _ = data.n_eff
    _ = data.k_eff


def test_flux_data():
    data = make_flux_data()
    _ = data.flux


def test_flux_time_data():
    data = make_flux_time_data()
    _ = data.flux


def test_diffraction_data():
    data = make_diffraction_data()
    _ = data.Etheta
    _ = data.Ephi
    _ = data.Er
    _ = data.Htheta
    _ = data.Hphi
    _ = data.Hr
    _ = data.orders_x
    _ = data.orders_y
    _ = data.f
    _ = data.ux
    _ = data.uy
    _ = data.angles
    _ = data.sim_size
    _ = data.bloch_vecs
    _ = data.amps
    _ = data.power
    _ = data.fields_spherical
    _ = data.fields_cartesian


def test_colocate():
    # TODO: can we colocate into regions where we dont store fields due to symmetry?
    # regular colocate
    data = make_field_data()
    _ = data.colocate(x=[+0.1, 0.5], y=[+0.1, 0.5], z=[+0.1, 0.5])

    # ignore coordinate
    _ = data.colocate(x=[+0.1, 0.5], y=None, z=[+0.1, 0.5])

    # data outside range of len(coord)==1 dimension
    data = make_mode_solver_data()
    with pytest.raises(DataError):
        _ = data.colocate(x=[+0.1, 0.5], y=1.0, z=[+0.1, 0.5])

    with pytest.raises(DataError):
        _ = data.colocate(x=[+0.1, 0.5], y=[1.0, 2.0], z=[+0.1, 0.5])


def test_time_reversed_copy():
    _ = make_field_data().time_reversed_copy
    _ = make_mode_solver_data().time_reversed_copy
    time_data = make_field_time_data()
    reversed_time_data = time_data.time_reversed_copy
    assert np.allclose(time_data.Ex.values, reversed_time_data.Ex.values[..., ::-1])
    assert np.allclose(time_data.Hx.values, -reversed_time_data.Hx.values[..., ::-1])


def _test_eq():
    data1 = make_flux_data()
    data2 = make_flux_data()
    data1.flux.data = np.ones_like(data1.flux.data)
    data2.flux.data = np.ones_like(data2.flux.data)
    data3 = make_flux_time_data_array()
    assert data1 == data2, "same data are not equal"
    data1.flux.data[0] = 1e12
    assert data1 != data2, "different data are equal"
    assert data1 != data3, "different data are equal"


def test_empty_array():
    coords = {"x": np.arange(10), "y": np.arange(10), "z": np.arange(10), "t": []}
    fields = {"Ex": td.ScalarFieldTimeDataArray(np.random.rand(10, 10, 10, 0), coords=coords)}
    monitor = td.FieldTimeMonitor(size=(1, 1, 1), fields=["Ex"], name="test")
    _ = td.FieldTimeData(
        monitor=monitor,
        symmetry=SIM.symmetry,
        symmetry_center=SIM.center,
        grid_expanded=SIM.discretize_monitor(monitor),
        **fields
    )


# NOTE: can remove this? lets not support empty tuple or list, use np.zeros()
def _test_empty_list():
    coords = {"x": np.arange(10), "y": np.arange(10), "z": np.arange(10), "t": []}
    fields = {"Ex": td.ScalarFieldTimeDataArray([], coords=coords)}
    monitor = td.FieldTimeMonitor(size=(1, 1, 1), fields=["Ex"], name="test")
    _ = td.FieldTimeData(
        monitor=monitor,
        symmetry=SIM.symmetry,
        symmetry_center=SIM.center,
        grid_expanded=SIM.discretize_monitor(monitor),
        **fields
    )


# NOTE: can remove this? lets not support empty tuple or list, use np.zeros()
def _test_empty_tuple():
    coords = {"x": np.arange(10), "y": np.arange(10), "z": np.arange(10), "t": []}
    fields = {"Ex": td.ScalarFieldTimeDataArray((), coords=coords)}
    monitor = td.FieldTimeMonitor(size=(1, 1, 1), fields=["Ex"], name="test")
    _ = td.FieldTimeData(
        monitor=monitor,
        symmetry=SIM.symmetry,
        symmetry_center=SIM.center,
        grid_expanded=SIM.discretize_monitor(monitor),
        **fields
    )


def test_empty_io(tmp_path):
    coords = {"x": np.arange(10), "y": np.arange(10), "z": np.arange(10), "t": []}
    fields = {"Ex": td.ScalarFieldTimeDataArray(np.random.rand(10, 10, 10, 0), coords=coords)}
    monitor = td.FieldTimeMonitor(size=(1, 1, 1), name="test", fields=["Ex"])
    field_data = td.FieldTimeData(
        monitor=monitor,
        symmetry=SIM.symmetry,
        symmetry_center=SIM.center,
        grid_expanded=SIM.discretize_monitor(monitor),
        **fields
    )
    field_data.to_file(str(tmp_path / "field_data.hdf5"))
    field_data = td.FieldTimeData.from_file(str(tmp_path / "field_data.hdf5"))
    assert field_data.Ex.size == 0


def test_mode_solver_plot_field():
    """Ensure we get a helpful error if trying to .plot_field with a ModeSolverData."""
    ms_data = make_mode_solver_data()
    with pytest.raises(DeprecationWarning):
        ms_data.plot_field(1, 2, 3, z=5, b=True)
    plt.close()


def test_field_data_symmetry_present():

    coords = {"x": np.arange(10), "y": np.arange(10), "z": np.arange(10), "t": []}
    fields = {"Ex": td.ScalarFieldTimeDataArray(np.random.rand(10, 10, 10, 0), coords=coords)}
    monitor = td.FieldTimeMonitor(size=(1, 1, 1), name="test", fields=["Ex"])

    # works if no symmetry specified
    _ = td.FieldTimeData(monitor=monitor, **fields)

    # fails if symmetry specified but missing symmetry center
    with pytest.raises(pydantic.ValidationError):
        _ = td.FieldTimeData(
            monitor=monitor,
            symmetry=(1, -1, 0),
            grid_expanded=SIM.discretize_monitor(monitor),
            **fields
        )

    # fails if symmetry specified but missing etended grid
    with pytest.raises(pydantic.ValidationError):
        _ = td.FieldTimeData(
            monitor=monitor, symmetry=(1, -1, 1), symmetry_center=(0, 0, 0), **fields
        )


def test_data_array_attrs():
    """Note, this is here because the attrs only get set when added to a pydantic model."""
    data = make_flux_data()
    assert data.flux.attrs, "data has no attrs"
    assert data.flux.f.attrs, "data coordinates have no attrs"


def test_data_array_json_warns(log_capture, tmp_path):
    data = make_flux_data()
    data.to_file(str(tmp_path / "flux.json"))
    assert_log_level(log_capture, "WARNING")


def test_data_array_hdf5_no_warnings(log_capture, tmp_path):
    data = make_flux_data()
    data.to_file(str(tmp_path / "flux.hdf5"))
    assert_log_level(log_capture, None)


def test_diffraction_data_use_medium():
    data = make_diffraction_data()
    data = data.copy(update=dict(medium=td.Medium(permittivity=4)))
    assert np.allclose(data.eta, np.real(td.ETA_0 / 2.0))


def test_mode_solver_data_sort():
    # test basic matching algorithm
    arr = np.array([[1, 2, 3], [6, 5, 4], [7, 9, 8]])
    pairs, values = ModeSolverData._find_closest_pairs(arr)
    assert np.all(pairs == [2, 0, 1])
    assert np.all(values == [3, 6, 9])

    # test sorting function
    # get smooth data
    data = make_mode_solver_data_smooth()
    # make it unsorted
    num_modes = len(data.Ex.coords["mode_index"])
    num_freqs = len(data.Ex.coords["f"])
    phases = 2 * np.pi * np.random.random((num_freqs, num_modes))
    unsorting = np.arange(num_modes) * np.ones((num_freqs, num_modes))
    unsorting = unsorting.astype(int)
    # we keep first, central, and last sorted
    for freq_id in range(1, num_freqs - 1):
        if freq_id != num_freqs // 2:
            unsorting[freq_id, :] = np.random.permutation(unsorting[freq_id, :])

    # unsort using sorting tool
    data_unsorted = data._reorder_modes(unsorting, phases, None)

    # sort back using all starting frequencies
    data_first = data_unsorted.overlap_sort(track_freq="lowest")
    data_last = data_unsorted.overlap_sort(track_freq="highest")
    data_center = data_unsorted.overlap_sort(track_freq="central")

    # check that sorted data coincides with original
    for data_sorted in [data_first, data_last, data_center]:
        for comp, field in data.field_components.items():
            assert np.allclose(np.abs(field), np.abs(data_sorted.field_components[comp]))
        assert np.allclose(data.n_complex, data_sorted.n_complex)
        assert np.allclose(data.grid_dual_correction, data_sorted.grid_dual_correction)
        assert np.allclose(data.grid_primal_correction, data_sorted.grid_primal_correction)

        # make sure neighboring frequencies are in phase
        data_1 = data._isel(f=[0])
        for i in range(1, num_freqs):
            data_2 = data._isel(f=[i])
            complex_amps = data_1.dot(data_2).data.ravel()
            data_1 = data_2
            assert np.all(np.abs(np.imag(complex_amps)) < 1e-15)


def test_mode_solver_numerical_grid_data():
    mode_data = make_mode_solver_data().symmetry_expanded_copy
    # _tangential_fields property applies the numerical correction and expands the symmetry
    tan_fields = mode_data._tangential_fields
    # Check that data is only slightly different
    for comp, field in mode_data.field_components.items():
        if comp in tan_fields.keys():
            max_diff = np.amax(np.abs(np.abs(field) - np.abs(tan_fields[comp])))
            max_diff /= np.amax(np.abs(field))
            assert 0.1 > max_diff > 0


def test_outer_dot():
    mode_data = make_mode_solver_data()
    field_data = make_field_data_2d()
    _ = mode_data.outer_dot(mode_data)
    _ = field_data.outer_dot(mode_data)
    _ = mode_data.outer_dot(field_data)
    _ = field_data.outer_dot(field_data)



"""Component imports for adjoint plugin. from tidy3d.plugins.adjoint.components import *"""

# import the jax version of tidy3d components
from .geometry import JaxBox  # , JaxPolySlab
from .medium import JaxMedium, JaxAnisotropicMedium, JaxCustomMedium
from .structure import JaxStructure
from .simulation import JaxSimulation
from .data.sim_data import JaxSimulationData
from .data.monitor_data import JaxModeData
from .data.dataset import JaxPermittivityDataset
from .data.data_array import JaxDataArray

__all__ = [
    "JaxBox",
    "JaxPolySlab",
    "JaxGeometryGroup",
    "JaxMedium",
    "JaxAnisotropicMedium",
    "JaxCustomMedium",
    "JaxStructure",
    "JaxSimulation",
    "JaxSimulationData",
    "JaxModeData",
    "JaxPermittivityDataset",
    "JaxDataArray",
]



"""Find resonances in time series data
"""

from typing import Tuple, List, Union
from functools import partial

import numpy as np
import scipy.linalg
import xarray as xr

from pydantic.v1 import Field, NonNegativeFloat, PositiveInt, validator

from ...log import log
from ...components.base import Tidy3dBaseModel
from ...components.types import ArrayFloat1D, ArrayComplex1D, ArrayComplex2D, ArrayComplex3D
from ...components.data.data_array import ScalarFieldTimeDataArray
from ...components.data.monitor_data import FieldTimeData
from ...constants import HERTZ
from ...exceptions import SetupError, ValidationError


INIT_NUM_FREQS = 200

TIME_STEP_RTOL = 1e-5

RCOND = 1e-4


# ResonanceData will be used internally
class ResonanceData(Tidy3dBaseModel):
    """Data class for storing objects computed while running the resonance finder."""

    eigvals: ArrayComplex1D = Field(..., title="Eigenvalues", description="Resonance eigenvalues.")
    complex_amplitudes: ArrayComplex1D = Field(
        None, title="Complex amplitudes", description="Complex resonance amplitudes"
    )
    errors: ArrayFloat1D = Field(
        None, title="Errors", description="Rough eigenvalue error estimate."
    )


class ResonanceFinder(Tidy3dBaseModel):
    """Tool that extracts resonance information from a time series of the form shown below.
    The resonance information consists of frequency :math:`f`, decay rate :math:`\\alpha`,
    Q factor :math:`Q = \\pi |f|/\\alpha`, amplitude :math:`a`, and phase :math:`\\phi`.

    Note
    ----
    .. math::

        f(t) = \\sum_k a_k e^{i \\phi_k} e^{-2 \\pi i f_k t - \\alpha_k t}

    Note
    ----
    We use the algorithm described in

    Vladimir A. Mandelshtam and Howard S. Taylor,
    "Harmonic inversion of time signals and its applications,"
    J. Chem. Phys. 107, 6756 (1997).

    Example
    -------
    >>> import numpy as np
    >>> from tidy3d.plugins.resonance import ResonanceFinder
    >>> t = np.linspace(0, 10000, 10000)
    >>> f1 = 2*np.pi*0.1 - 1j*0.002
    >>> f2 = 2*np.pi*0.2 - 1j*0.0005
    >>> sig = 2*np.exp(-1j*f1*t) + 3*1j*np.exp(-1j*f2*t)
    >>> resfinder = ResonanceFinder(freq_window=(0.05, 0.25))
    >>> resdata = resfinder.run_raw_signal(signal=sig, time_step=1)
    >>> resdata.to_dataframe()

    """

    freq_window: Tuple[float, float] = Field(
        ...,
        title="Window ``[fmin, fmax]``",
        description="Window ``[fmin, fmax]`` for the initial frequencies. "
        "The resonance finder is initialized with an even grid of frequencies between "
        "fmin and fmax. The resonance finder then iteratively optimizes and prunes these "
        "frequencies. Note that frequencies outside this window may be returned. "
        "A narrow frequency window that only contains a few resonances may give enhanced "
        "accuracy compared to a broad frequency window with many resonances.",
        units=HERTZ,
    )

    init_num_freqs: PositiveInt = Field(
        INIT_NUM_FREQS,
        title="Initial number of frequencies.",
        description="Number of frequencies with which the resonance finder is initialized. "
        "The resonance finder then iteratively optimizes and prunes these frequencies. "
        "The number of frequencies returned will be less than ``init_num_freqs``. "
        "Making this larger can find more resonances, while making it smaller can speed up "
        "the resonance finder.",
    )

    rcond: NonNegativeFloat = Field(
        RCOND,
        title="Cutoff for eigenvalues",
        description="Cutoff for eigenvalues, relative to the largest eigenvalue. "
        "The resonance finder solves a generalized eigenvalue problem of the form "
        ":math:`Ax = \\lambda B x`. If B has small eigenvalues, this is poorly conditioned, "
        "so we eliminate eigenvalues of B less than ``rcond`` times the largest eigenvalue. "
        "Making this closer to zero will typically return more resonances.",
    )

    @validator("freq_window", always=True)
    def _check_freq_window(cls, val):
        """Validate ``freq_window``"""
        if val[1] < val[0]:
            raise ValidationError(
                "The value of 'freq_window[1]' is less than the value of 'freq_window[0]'."
            )
        return val

    def run(self, signals: Union[FieldTimeData, Tuple[FieldTimeData, ...]]) -> xr.Dataset:
        """Finds resonances in a :class:`.FieldTimeData` or a Tuple of such.
        The time coordinates must be uniformly spaced, and the spacing must be the same
        across all supplied data. The resonance finder runs on the sum of the
        ``Ex``, ``Ey``, and ``Ez`` fields of all the supplied data,
        unless no electric fields are present at all, in which case it runs on the sum of the
        ``Hx``, ``Hy``, and ``Hz`` fields.
        Note that the signal should start after the sources have turned off.

        Parameters
        ----------
        signals: :class:`.FieldTimeData`
            Data to search for resonances

        Returns
        -------
        xr.Dataset
            Dataset containing the decay rate, Q, amplitude, phase, and estimation error
            of the resonances as a function of frequency. Modes with low Q, small
            amplitude, or high estimation error are likely to be spurious.
        """
        if isinstance(signals, FieldTimeData):
            signals = (signals,)
        if len(signals) == 0:
            raise SetupError("No 'FieldTimeData' supplied.")
        return self.run_scalar_field_time(self._aggregate_field_time(signals))

    def run_scalar_field_time(self, signal: ScalarFieldTimeDataArray) -> xr.Dataset:
        """Finds resonances in a :class:`.ScalarFieldTimeDataArray`.
        The time coordinates must be uniformly spaced to use the resonance finder;
        the time step is calculated automatically.
        Note that the signal should start after the sources have turned off.

        Parameters
        ----------
        signal : :class:`.ScalarFieldTimeDataArray`
            Time series to search for resonances

        Returns
        -------
        xr.Dataset
            Dataset containing the decay rate, Q, amplitude, phase, and estimation error
            of the resonances as a function of frequency. Modes with low Q, small
            amplitude, or high estimation error are likely to be spurious.
        """
        signal, dt = self._validate_scalar_field_time(signal)
        return self.run_raw_signal(signal, dt)

    def run_raw_signal(self, signal: List[complex], time_step: float) -> xr.Dataset:
        """Finds resonances in a time series.
        Note that the signal should start after the sources have turned off.

        Parameters
        ----------
        signal : List[complex]
            One-dimensional array holding the complex-valued time series data
            to search for resonances.
        time_step : float
            Time step / sampling rate of the data (in seconds).

        Returns
        -------
        xr.Dataset
            Dataset containing the decay rate, Q, amplitude, phase, and estimation error
            of the resonances as a function of frequency. Modes with low Q, small
            amplitude, or high estimation error are likely to be spurious.
        """
        signal = np.array(signal)
        if len(signal.shape) != 1:
            raise SetupError("The input signal should only have one dimension.")
        fmin, fmax = self.freq_window
        nfreqs = self.init_num_freqs
        log.info(f"\tRunning ResonanceFinder (nfreqs = {nfreqs})")
        omegas = np.linspace(
            time_step * fmin * 2 * np.pi,
            time_step * fmax * 2 * np.pi,
            nfreqs,
        )
        eigvals = np.exp(-1j * omegas)
        resdata = ResonanceData(eigvals=eigvals)
        resdata = self._iterate(signal, resdata)
        prev_num_eigvals = len(resdata.eigvals)
        for _ in range(1 + 2 * nfreqs):
            resdata = self._iterate(signal, resdata)
            new_num_eigvals = len(resdata.eigvals)
            if new_num_eigvals == prev_num_eigvals:
                log.info(f"\tCompleted ResonanceFinder (nfreqs = {new_num_eigvals})")
                break
            log.info(f"\tIterated ResonanceFinder (nfreqs = {new_num_eigvals})")
            prev_num_eigvals = new_num_eigvals

        resonances = self._get_resonance_info(data=resdata, time_step=time_step)

        return resonances.sortby("freq")

    def _validate_scalar_field_time(
        self, signal: ScalarFieldTimeDataArray
    ) -> Tuple[ArrayComplex1D, float]:
        """Validates a :class:`.ScalarFieldTimeDataArray` and returns the time step
        as well as underlying data array."""
        dts = np.diff(signal.t)
        dt = dts[0]
        if not np.allclose(dts, dt, rtol=1e-5, atol=0):
            raise SetupError(
                "The time coordinates in the supplied "
                "'ScalarFieldTimeDataArray' must be uniformly spaced to use "
                "the resonance finder."
            )
        for dim in signal.dims:
            if dim != "t" and len(signal.coords[dim]) > 1:
                raise SetupError(
                    "The length of each coordinate besides 't' of the supplied "
                    "'ScalarFieldTimeDataArray' must be equal to one."
                )
        return np.squeeze(signal.data), dt

    def _aggregate_field_time_comps(
        self, signals: Tuple[FieldTimeData, ...], comps
    ) -> ScalarFieldTimeDataArray:
        """Aggregates the given components from several :class:`.FieldTimeData`."""
        total_signal = None
        dt = -1
        coords = dict(x=[0], y=[0], z=[0], t=[0])

        for sig_field in signals:
            for comp in comps:
                field_comp = getattr(sig_field, comp, None)
                if field_comp is not None:
                    curr_signal, curr_dt = self._validate_scalar_field_time(field_comp)
                    if dt == -1:
                        dt = curr_dt
                        coords["t"] = field_comp.t
                    elif abs(curr_dt - dt) > TIME_STEP_RTOL:
                        raise SetupError(
                            "The time steps in the supplied 'FieldTimeData' must all be equal."
                        )
                    if total_signal is None:
                        total_signal = np.zeros(len(curr_signal), dtype=complex)
                    elif len(curr_signal) != len(total_signal):
                        raise SetupError(
                            "The length of the supplied 'FieldTimeData' must all be equal."
                        )
                    total_signal += curr_signal

        if total_signal is None:
            return None
        return ScalarFieldTimeDataArray(
            np.reshape(total_signal, (1, 1, 1, len(total_signal))), coords=coords
        )

    def _aggregate_field_time(
        self, signals: Union[FieldTimeData, Tuple[FieldTimeData, ...]]
    ) -> ScalarFieldTimeDataArray:
        """Aggregates several :class:`.FieldTimeData` into a single
        :class:`.ScalarFieldTimeDataArray`."""
        electric_components = ["Ex", "Ey", "Ez"]
        magnetic_components = ["Hx", "Hy", "Hz"]

        # first check electric components
        electric_data = self._aggregate_field_time_comps(signals, electric_components)
        if electric_data is not None:
            return electric_data
        # now check magnetic components
        magnetic_data = self._aggregate_field_time_comps(signals, magnetic_components)

        if magnetic_data is None:
            raise SetupError("No field components supplied.")
        return magnetic_data

    def _get_resonance_info(self, data: ResonanceData, time_step: float) -> xr.Dataset:
        """Extracts resonance information from the ResonanceData, and packages it as
        an xr.Dataset."""
        complex_omegas = 1j * np.log(data.eigvals)
        freqs = np.real(complex_omegas / (2 * np.pi)) / time_step
        decays = -np.imag(complex_omegas) / time_step
        q_factors = np.pi * np.abs(freqs) / decays
        amplitudes = np.abs(data.complex_amplitudes)
        phases = np.angle(data.complex_amplitudes)
        errors = np.array(data.errors)

        coords = {"freq": freqs}
        keys = ("decay", "Q", "amplitude", "phase", "error")
        vals = (decays, q_factors, amplitudes, phases, errors)

        data_arrays = tuple(map(partial(xr.DataArray, coords=coords), vals))
        return xr.Dataset(dict(zip(keys, data_arrays)))

    def _evaluate_matrices(self, signal: ArrayComplex1D, eigvals: ArrayComplex1D) -> ArrayComplex3D:
        """Compute the evolution matrices"""
        half_len = int(len(signal) / 2) - 2
        nfreqs = len(eigvals)
        zvals = eigvals / np.abs(eigvals)

        zinvl = np.exp(1j * np.outer(np.log(zvals) * 1j, np.arange(2 * half_len + 1)))

        prefactor = np.zeros((nfreqs, nfreqs), dtype=complex)

        prefactor[~np.eye(nfreqs, dtype=bool)] = np.reciprocal(
            np.subtract.outer(zvals, zvals)[~np.eye(nfreqs, dtype=bool)]
        )

        u_matrices = np.zeros((3, nfreqs, nfreqs), dtype=complex)
        for pval in range(3):

            u_matrices[pval, :, :] = prefactor * (
                np.outer(zvals, zinvl[:, : half_len + 1] @ signal[pval:][: half_len + 1])
                + np.outer(
                    zinvl[:, :half_len] @ signal[pval:][half_len + 1 : 2 * half_len + 1],
                    zinvl[:, half_len],
                )
            )
            u_matrices[pval, :, :] += u_matrices[pval, :, :].T

            for i in range(nfreqs):
                u_matrices[pval, i, i] = np.dot(
                    zinvl[i, : 2 * half_len + 1],
                    np.concatenate((np.arange(1, half_len + 2), np.arange(half_len, 0, -1)))
                    * signal[pval:][: 2 * half_len + 1],
                )

        return u_matrices

    def _gram_schmidt(self, a_matrix: ArrayComplex2D) -> ArrayComplex2D:
        """Perform the Gram-Schmidt process on the columns of a matrix."""
        new_a_matrix = np.zeros(a_matrix.shape, dtype=complex)
        for i in range(new_a_matrix.shape[1]):
            new_a_matrix[:, i] = a_matrix[:, i]
            for j in range(i):
                new_a_matrix[:, i] -= new_a_matrix[:, j] * np.dot(
                    new_a_matrix[:, j], a_matrix[:, i]
                )
            new_a_matrix[:, i] /= np.sqrt(np.dot(new_a_matrix[:, i], new_a_matrix[:, i]))
        return new_a_matrix

    def _solve_gen_eig_prob(
        self, a_matrix: ArrayComplex2D, b_matrix: ArrayComplex2D, rcond: float
    ) -> Tuple[ArrayComplex1D, ArrayComplex2D]:
        """Solve a generalized eigenvalue problem of the form

        .. math::
        Av = \\lambda B v

        We make B nonsingular by eliminating eigenvalues smaller
        than ```rcond`` times the largest eigenvalue. This method is used in
        Michael R. Wall and Daniel Neuhauser,
        "Extraction, through filter-diagonalization, of general quantum eigenvalues
        or classical normal mode frequencies of a short-time segment of a signal.
        I. Theory and application to a quantum-dynamics model,"
        J. Chem. Phys. 102, 8001 (1995).
        """
        eigvals_b, eigvecs_b = scipy.linalg.eig(b_matrix)
        large_inds = np.abs(eigvals_b) > rcond * np.amax(np.abs(eigvals_b))
        eigvals_b = eigvals_b[large_inds]
        eigvecs_b = eigvecs_b[:, large_inds]
        eigvecs_b = self._gram_schmidt(eigvecs_b)

        change_basis = eigvecs_b @ np.diag(1 / np.sqrt(eigvals_b))
        new_a_matrix = change_basis.T @ a_matrix @ change_basis
        eigvals, eigvecs_a = scipy.linalg.eig(new_a_matrix)
        eigvecs_a = self._gram_schmidt(eigvecs_a)

        eigvecs = change_basis @ eigvecs_a

        return eigvals, eigvecs

    def _find_amplitudes(
        self,
        signal: ArrayComplex1D,
        prev_eigvals: ArrayComplex1D,
        eigvals: ArrayComplex1D,
        eigvecs: ArrayComplex2D,
    ) -> ArrayComplex1D:
        """Compute the resonance amplitudes."""
        half_len = int(len(signal) / 2) - 2
        nfreqs = len(eigvals)
        zvals = prev_eigvals / np.abs(prev_eigvals)
        amplitudes = np.zeros(nfreqs, dtype=complex)

        zinvl = np.exp(1j * np.outer(np.log(zvals) * 1j, np.arange(half_len + 1)))
        factor = zinvl @ signal[: half_len + 1]

        for k in range(nfreqs):
            amplitudes[k] = np.square(eigvecs[:, k] @ factor)

        return amplitudes

    def _find_errors(
        self,
        eigvals: ArrayComplex1D,
        u_matrices: ArrayComplex3D,
        eigvecs: ArrayComplex2D,
    ) -> ArrayComplex1D:
        """Estimate the eigenvalue error."""
        nfreqs = len(eigvals)
        errors = np.zeros(nfreqs)
        for k in range(nfreqs):
            errors[k] = np.linalg.norm(
                (u_matrices[2] - (eigvals[k] ** 2) * u_matrices[0]) @ eigvecs[:, k]
            )
        return errors

    def _iterate(self, signal: ArrayComplex1D, prev_resdata: ResonanceData) -> ResonanceData:
        """Run a single iteration of the resonance finder."""
        prev_eigvals = prev_resdata.eigvals

        u_matrices = self._evaluate_matrices(signal, prev_eigvals)

        eigvals, eigvecs = self._solve_gen_eig_prob(u_matrices[1], u_matrices[0], self.rcond)

        errors = self._find_errors(eigvals, u_matrices, eigvecs)

        amplitudes = self._find_amplitudes(signal, prev_eigvals, eigvals, eigvecs)

        return ResonanceData(eigvals=eigvals, errors=errors, complex_amplitudes=amplitudes)



import pytest
import numpy as np
import os
import sys
from memory_profiler import profile
import matplotlib.pylab as plt
import time
import jax

import jax.numpy as jnp
from ..utils import run_emulated

import tidy3d as td
import tidy3d.plugins.adjoint as tda
from tidy3d.plugins.adjoint.web import run_local as run

import cProfile

# name of the output monitor used in tests
MNT_NAME = "field"

WAVELENGTH = 1.0
BOX_LENGTH = 1.0
SPACE = 1.0
FREQ0 = td.C_0 / WAVELENGTH
N_SIDE = 15

surface_area = 6 * (N_SIDE * BOX_LENGTH**2)


def make_sim(box_length) -> tda.JaxSimulation:
    """Construt a simulation out of some input parameters."""

    num_structures = int(N_SIDE**2)
    sim_length = N_SIDE * (BOX_LENGTH + SPACE)
    med = tda.JaxMedium(permittivity=2.0)

    xmin = (-sim_length + SPACE + BOX_LENGTH) / 2
    ymin = (-sim_length + SPACE + BOX_LENGTH) / 2

    structures = []
    for i in range(N_SIDE):
        x0 = xmin + i * (SPACE + BOX_LENGTH)
        for j in range(N_SIDE):
            y0 = ymin + j * (SPACE + BOX_LENGTH)
            box = tda.JaxBox(size=(box_length, box_length, box_length), center=(x0, y0, 0))
            struct = tda.JaxStructure(geometry=box, medium=med)
            structures.append(struct)

    # ModeMonitors
    mnt = td.FieldMonitor(
        size=(0, 0, 0),
        center=(0, 0, (BOX_LENGTH + SPACE / 2) / 2),
        freqs=[FREQ0],
        name=MNT_NAME,
    )

    sim = tda.JaxSimulation(
        size=(sim_length, sim_length, 2 * BOX_LENGTH + 2 * SPACE),
        run_time=1e-12,
        grid_spec=td.GridSpec(wavelength=1.0),
        input_structures=structures,
        output_monitors=(mnt,),
        normalize_index=None,
    )

    return sim


@pytest.fixture
def use_emulated_run(monkeypatch):
    """If this fixture is used, the `tests.utils.run_emulated` function is used for simulation."""
    import tidy3d.plugins.adjoint.web as adjoint_web

    monkeypatch.setattr(adjoint_web, "tidy3d_run_fn", run_emulated)


@profile
def test_large_custom_medium(use_emulated_run):
    def f(box_length):
        sim = make_sim(box_length)
        new_size = list(sim.size)
        new_size[0] += 1.0
        sim2 = sim.updated_copy(size=new_size)
        sim_data = run(sim2, task_name="test")
        fld = sim_data[MNT_NAME].Ez
        fld2 = fld.copy()
        fld2 = fld2.values
        return jnp.sum(jnp.abs(jnp.array(fld2)))

    with cProfile.Profile() as pr:
        grad_f = jax.grad(f)
        df_eps_values = grad_f(BOX_LENGTH)
        pr.print_stats(sort="cumtime")
        pr.dump_stats("results.prof")

    # with cProfile.Profile() as pr:
    #     res = f(BOX_LENGTH)
    #     res2 = f(BOX_LENGTH)
    #     pr.print_stats(sort="cumtime")
    #     pr.dump_stats("results.prof")



"""Tests file export and loading."""
import os
import json

import pytest
import numpy as np
from time import time
import dill as pickle


from tidy3d import __version__
import tidy3d as td
from tidy3d.components.base import DATA_ARRAY_MAP
from ..utils import SIM_FULL as SIM
from ..utils import SIM_MONITORS as SIM2
from ..test_data.test_monitor_data import make_flux_data
from ..test_data.test_sim_data import make_sim_data
from tidy3d.components.data.sim_data import DATA_TYPE_MAP

# Store an example of every minor release simulation to test updater in the future
SIM_DIR = "tests/sims"


def set_datasets_to_none(sim):
    sim_dict = sim.dict()
    for src in sim_dict["sources"]:
        if src["type"] == "CustomFieldSource":
            src["field_dataset"] = None
        elif src["type"] == "CustomCurrentSource":
            src["current_dataset"] = None
        if src["source_time"]["type"] == "CustomSourceTime":
            src["source_time"]["source_time_dataset"] = None
    for structure in sim_dict["structures"]:
        if structure["geometry"]["type"] == "TriangleMesh":
            structure["geometry"]["mesh_dataset"] = None
        if structure["geometry"]["type"] == "GeometryGroup":
            for geometry in structure["geometry"]["geometries"]:
                if geometry["type"] == "TriangleMesh":
                    geometry["mesh_dataset"] = None
        if "Custom" in structure["medium"]["type"]:
            if structure["medium"]["type"] == "CustomMedium":
                structure["medium"]["eps_dataset"] = None
            elif structure["medium"]["type"] == "CustomPoleResidue":
                structure["medium"]["poles"] = []
            else:
                structure["medium"]["coeffs"] = []
    return td.Simulation.parse_obj(sim_dict)


def test_simulation_load_export():
    major, minor, patch = __version__.split(".")
    path = os.path.join(SIM_DIR, f"simulation_{major}_{minor}_{patch}.json")
    SIM.to_file(path)
    SIM2 = td.Simulation.from_file(path)
    assert set_datasets_to_none(SIM) == SIM2, "original and loaded simulations are not the same"


def test_simulation_load_export_yaml(tmp_path):
    path = str(tmp_path / "simulation.yaml")
    SIM.to_file(path)
    SIM2 = td.Simulation.from_file(path)
    assert set_datasets_to_none(SIM) == SIM2, "original and loaded simulations are not the same"


def test_component_load_export(tmp_path):
    path = str(tmp_path / "medium.json")
    td.Medium().to_file(path)
    M2 = td.Medium.from_file(path)
    assert td.Medium() == M2, "original and loaded medium are not the same"


def test_component_load_export_yaml(tmp_path):
    path = str(tmp_path / "medium.yaml")
    td.Medium().to_file(path)
    M2 = td.Medium.from_file(path)
    assert td.Medium() == M2, "original and loaded medium are not the same"


def test_simulation_load_export_hdf5(tmp_path):
    path = str(tmp_path / "simulation.hdf5")
    SIM.to_file(path)
    SIM2 = td.Simulation.from_file(path)
    assert SIM == SIM2, "original and loaded simulations are not the same"


def test_simulation_load_export_hdf5_gz(tmp_path):
    path = str(tmp_path / "simulation.hdf5.gz")
    SIM.to_file(path)
    SIM2 = td.Simulation.from_file(path)
    assert SIM == SIM2, "original and loaded simulations are not the same"


def test_simulation_load_export_hdf5_explicit(tmp_path):
    path = str(tmp_path / "simulation.hdf5")
    SIM.to_hdf5(path)
    SIM2 = td.Simulation.from_hdf5(path)
    assert SIM == SIM2, "original and loaded simulations are not the same"


def test_simulation_load_export_hdf5_gz_explicit(tmp_path):
    path = str(tmp_path / "simulation.hdf5.gz")
    SIM.to_hdf5_gz(path)
    SIM2 = td.Simulation.from_hdf5_gz(path)
    assert SIM == SIM2, "original and loaded simulations are not the same"


def test_simulation_load_export_pckl(tmp_path):
    path = str(tmp_path / "simulation.pckl")
    with open(path, "wb") as pickle_file:
        pickle.dump(SIM, pickle_file)
    with open(path, "rb") as pickle_file:
        SIM2 = pickle.load(pickle_file)
    assert SIM == SIM2, "original and loaded simulations are not the same"


def test_simulation_preserve_types(tmp_path):
    """Test that all re-loaded components have the same types."""
    path = str(tmp_path / "simulation.json")
    SIM.to_file(path)
    sim_2 = td.Simulation.from_file(path)
    assert set_datasets_to_none(SIM) == sim_2

    M_types = [type(s.medium) for s in sim_2.structures]
    for M in (td.Medium, td.PoleResidue, td.Lorentz, td.Sellmeier, td.Debye):
        assert M in M_types

    G_types = [type(s.geometry) for s in sim_2.structures]
    for G in (td.Box, td.Sphere, td.Cylinder, td.PolySlab):
        assert G in G_types

    S_types = [type(s) for s in sim_2.sources]
    for S in (td.UniformCurrentSource, td.PlaneWave, td.GaussianBeam):
        assert S in S_types

    M_types = [type(m) for m in sim_2.monitors]
    for M in (
        td.FieldMonitor,
        td.FieldTimeMonitor,
        td.ModeMonitor,
        td.FluxMonitor,
        td.FluxTimeMonitor,
    ):
        assert M in M_types


def test_1a_simulation_load_export2(tmp_path):
    path = str(tmp_path / "simulation.json")
    SIM2.to_file(path)
    SIM3 = td.Simulation.from_file(path)
    assert SIM2 == SIM3, "original and loaded simulations are not the same"


def test_validation_speed(tmp_path):
    sizes_bytes = []
    times_sec = []
    path = str(tmp_path / "simulation.json")

    _ = SIM
    N_tests = 10

    # adjust as needed, keeping small to speed tests up
    num_structures = np.logspace(0, 2, N_tests).astype(int)

    for n in num_structures:
        new_structures = []
        for i in range(n):
            new_structure = SIM.structures[0].copy(update={"name": str(i)})
            new_structures.append(new_structure)
        S = SIM.copy(update=dict(structures=new_structures))

        S.to_file(path)
        time_start = time()
        _S = td.Simulation.from_file(path)
        time_validate = time() - time_start
        times_sec.append(time_validate)
        assert set_datasets_to_none(S) == _S

        size = os.path.getsize(path)
        sizes_bytes.append(size)

        print(f"{n} structures \t {size:.1e} bytes \t {time_validate:.1f} seconds to validate")


SIM_FILES = [os.path.join(SIM_DIR, file) for file in os.listdir(SIM_DIR)]


@pytest.mark.parametrize("sim_file", SIM_FILES)
def test_simulation_updater(sim_file):
    """Test that all simulations in ``SIM_DIR`` can be updated to current version and loaded."""
    sim_updated = td.Simulation.from_file(sim_file)
    assert sim_updated.version == __version__, "Simulation not converted properly"

    # just make sure the loaded sim does something properly using this version
    sim_updated.grid


def test_yaml(tmp_path):
    path = str(tmp_path / "simulation.json")
    SIM.to_file(path)
    sim = td.Simulation.from_file(path)
    path1 = str(tmp_path / "simulation.yaml")
    sim.to_yaml(path1)
    sim1 = td.Simulation.from_yaml(path1)
    assert sim1 == sim


def test_to_json_data():
    """Tests that the json string with data in separate file behaves correctly."""

    # type saved in the combined json file?
    data = make_flux_data()
    json_dict = json.loads(data._json_string)
    assert json_dict["flux"] in DATA_ARRAY_MAP


def test_to_hdf5_group_path_sim_data(tmp_path):
    """Tests that the json string with data in separate file behaves correctly in SimulationData."""

    # type saved in the combined json file?
    sim_data = make_sim_data()
    FNAME = str(tmp_path / "sim_data.hdf5")
    sim_data.to_file(fname=FNAME)

    for i, monitor in enumerate(sim_data.simulation.monitors):
        group_name = sim_data.get_tuple_group_name(index=i)
        group_path = f"/data/{group_name}"
        MntDataType = DATA_TYPE_MAP[type(monitor)]
        mnt_data = MntDataType.from_file(fname=FNAME, group_path=group_path)
        assert mnt_data == sim_data.monitor_data[monitor.name]


def test_none_hdf5(tmp_path):
    """Tests that values of None where None is not the default are loaded correctly."""

    sim = td.Simulation(
        size=(1, 1, 1),
        grid_spec=td.GridSpec(wavelength=1.0),
        run_time=1e-12,
        normalize_index=None,
    )

    assert sim.normalize_index is None, "'normalize_index' of 'None' not initialized correctly."

    fname = str(tmp_path / "sim_none.hdf5")
    sim.to_file(fname)
    sim2 = td.Simulation.from_file(fname)

    assert sim2.normalize_index is None, "'normalize_index' of 'None' not loaded correctly."


def test_group_name_tuple():
    """Test conversion of group names and tuples."""

    tidy = td.Medium()
    tuple_values = ["Something", "Another thing", "Something different entirely"]
    test_dict = tidy.tuple_to_dict(tuple_values=tuple_values)

    # make sure we can pick out the index from the dictionary keys and get the correct key
    for true_index, key_name in enumerate(test_dict.keys()):
        index = tidy.get_tuple_index(key_name=key_name)
        assert index == true_index
        group_name = tidy.get_tuple_group_name(index=index)
        assert group_name == key_name



"""Utilities for geometry manipulation."""
from __future__ import annotations
from typing import Union, Tuple

from ..types import Axis, PlanePosition, Shapely
from ...exceptions import Tidy3dError

from . import base
from . import primitives
from . import polyslab
from . import mesh

GeometryType = Union[
    base.Box,
    base.ClipOperation,
    base.GeometryGroup,
    primitives.Sphere,
    primitives.Cylinder,
    polyslab.PolySlab,
    polyslab.ComplexPolySlabBase,
    mesh.TriangleMesh,
]


def flatten_groups(*geometries: GeometryType) -> GeometryType:
    """Iterates over all geometries, flattening groups and unions.

    Parameters
    ----------
    *geometries: GeometryType
        Geometries to flatten.

    Returns
    -------
    :class:`Geometry`
        Geometries after flattening groups and unions."""
    for geometry in geometries:
        if isinstance(geometry, base.GeometryGroup):
            yield from flatten_groups(*geometry.geometries)
        elif isinstance(geometry, base.ClipOperation) and geometry.operation == "union":
            yield from flatten_groups(geometry.geometry_a, geometry.geometry_b)
        else:
            yield geometry


def traverse_geometries(geometry: GeometryType) -> GeometryType:
    """Iterator over all geometries within the given geometry.

    Iterates over groups and clip operations within the given geometry, yielding each one.

    Parameters
    ----------
    geometry: GeometryType
        Base geometry to start iteration.

    Returns
    -------
    :class:`Geometry`
        Geometries within the base geometry.
    """
    if isinstance(geometry, base.GeometryGroup):
        for g in geometry.geometries:
            yield from traverse_geometries(g)
    elif isinstance(geometry, base.ClipOperation):
        yield from traverse_geometries(geometry.geometry_a)
        yield from traverse_geometries(geometry.geometry_b)
    yield geometry


# pylint:disable=too-many-arguments
def from_shapely(
    shape: Shapely,
    axis: Axis,
    slab_bounds: Tuple[float, float],
    dilation: float = 0.0,
    sidewall_angle: float = 0,
    reference_plane: PlanePosition = "middle",
) -> base.Geometry:
    """Convert a shapely primitive into a geometry instance by extrusion.

    Parameters
    ----------
    shape : shapely.geometry.base.BaseGeometry
        Shapely primitive to be converted. It must be a linear ring, a polygon or a collection
        of any of those.
    axis : int
        Integer index defining the extrusion axis: 0 (x), 1 (y), or 2 (z).
    slab_bounds: Tuple[float, float]
        Minimal and maximal positions of the extruded slab along ``axis``.
    dilation : float
        Dilation of the polygon in the base by shifting each edge along its normal outwards
        direction by a distance; a negative value corresponds to erosion.
    sidewall_angle : float = 0
        Angle of the extrusion sidewalls, away from the vertical direction, in radians. Positive
        (negative) values result in slabs larger (smaller) at the base than at the top.
    reference_plane : PlanePosition = "middle"
        Reference position of the (dilated/eroded) polygons along the slab axis. One of
        ``"middle"`` (polygons correspond to the center of the slab bounds), ``"bottom"``
        (minimal slab bound position), or ``"top"`` (maximal slab bound position). This value
        has no effect if ``sidewall_angle == 0``.

    Returns
    -------
    :class:`Geometry`
        Geometry extruded from the 2D data.
    """
    if shape.geom_type == "LinearRing":
        if sidewall_angle == 0:
            return polyslab.PolySlab(
                vertices=shape.coords[:-1],
                axis=axis,
                slab_bounds=slab_bounds,
                dilation=dilation,
                reference_plane=reference_plane,
            )
        group = polyslab.ComplexPolySlabBase(
            vertices=shape.coords[:-1],
            axis=axis,
            slab_bounds=slab_bounds,
            dilation=dilation,
            sidewall_angle=sidewall_angle,
            reference_plane=reference_plane,
        ).geometry_group
        return group.geometries[0] if len(group.geometries) == 1 else group

    if shape.geom_type == "Polygon":
        exterior = from_shapely(
            shape.exterior, axis, slab_bounds, dilation, sidewall_angle, reference_plane
        )
        interior = [
            from_shapely(hole, axis, slab_bounds, -dilation, -sidewall_angle, reference_plane)
            for hole in shape.interiors
        ]
        if len(interior) == 0:
            return exterior
        interior = interior[0] if len(interior) == 1 else base.GeometryGroup(geometries=interior)
        return base.ClipOperation(operation="difference", geometry_a=exterior, geometry_b=interior)

    if shape.geom_type in {"MultiPolygon", "GeometryCollection"}:
        return base.GeometryGroup(
            geometries=[
                from_shapely(geo, axis, slab_bounds, dilation, sidewall_angle, reference_plane)
                for geo in shape.geoms
            ]
        )

    raise Tidy3dError(f"Shape {shape} cannot be converted to Geometry.")



"""Defines geometric transformation classes"""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Union

import pydantic.v1 as pd
import numpy as np

from .base import Tidy3dBaseModel, cached_property
from .types import Coordinate, TensorReal, ArrayFloat2D
from ..constants import RADIAN


class AbstractRotation(ABC, Tidy3dBaseModel):
    """Abstract rotation of vectors and tensors."""

    @cached_property
    @abstractmethod
    def matrix(self) -> TensorReal:
        """Rotation matrix."""

    @cached_property
    @abstractmethod
    def isidentity(self) -> bool:
        """Check whether rotation is identity."""

    def rotate_vector(self, vector: ArrayFloat2D) -> ArrayFloat2D:
        """Rotate a vector/point or a list of vectors/points.

        Parameters
        ----------
        points : ArrayLike[float]
            Array of shape ``(3, ...)``.

        Returns
        -------
        Coordinate
            Rotated vector.
        """

        if self.isidentity:
            return vector

        if len(vector.shape) == 1:
            return self.matrix @ vector

        return np.tensordot(self.matrix, vector, axes=1)

    def rotate_tensor(self, tensor: TensorReal) -> TensorReal:
        """Rotate a tensor.

        Parameters
        ----------
        tensor : ArrayLike[float]
            Array of shape ``(3, 3)``.

        Returns
        -------
        TensorReal
            Rotated tensor.
        """

        if self.isidentity:
            return tensor

        return np.matmul(self.matrix, np.matmul(tensor, self.matrix.T))


class RotationAroundAxis(AbstractRotation):
    """Rotation of vectors and tensors around a given vector."""

    axis: Coordinate = pd.Field(
        [1, 0, 0],
        title="Axis of Rotation",
        description="A vector that specifies the axis of rotation.",
    )

    angle: float = pd.Field(
        0.0,
        title="Angle of Rotation",
        description="Angle of rotation in radians.",
        units=RADIAN,
    )

    @cached_property
    def isidentity(self) -> bool:
        """Check whether rotation is identity."""

        return np.isclose(self.angle % (2 * np.pi), 0)

    @cached_property
    def matrix(self) -> TensorReal:
        """Rotation matrix."""

        if self.isidentity:
            return np.eye(3)

        n = self.axis / np.linalg.norm(self.axis)
        c = np.cos(self.angle)
        s = np.sin(self.angle)
        R = np.zeros((3, 3))
        tan_dim = [[1, 2], [2, 0], [0, 1]]

        for dim in range(3):
            R[dim, dim] = c + n[dim] ** 2 * (1 - c)
            R[dim, tan_dim[dim][0]] = n[dim] * n[tan_dim[dim][0]] * (1 - c) - n[tan_dim[dim][1]] * s
            R[dim, tan_dim[dim][1]] = n[dim] * n[tan_dim[dim][1]] * (1 - c) + n[tan_dim[dim][0]] * s

        return R


RotationType = Union[RotationAroundAxis]



"""File compression utilities"""

import gzip
import shutil


def compress_file_to_gzip(input_file, output_gz_file):
    """
    Compresses a file using gzip.

    Args:
        input_file (str): The path of the input file.
        output_gz_file (str): The path of the output gzip file.
    """
    with open(input_file, "rb") as file_in:
        with gzip.open(output_gz_file, "wb") as file_out:
            shutil.copyfileobj(file_in, file_out)


def extract_gzip_file(input_gz_file, output_file):
    """
    Extract a gzip file.

    Args:
        input_gz_file (str): The path of the gzip input file.
        output_file (str): The path of the output file.
    """
    with gzip.open(input_gz_file, "rb") as file_in:
        with open(output_file, "wb") as file_out:
            shutil.copyfileobj(file_in, file_out)



"""Tests tidy3d/components/data/data_array.py"""
import numpy as np
from typing import Tuple, List

import tidy3d as td

np.random.seed(4)

STRUCTURES = [
    td.Structure(
        geometry=td.Box(size=(1, td.inf, 1)), medium=td.material_library["cSi"]["SalzbergVilla1957"]
    )
]
SIZE_3D = (2, 4, 5)
SIZE_2D = list(SIZE_3D)
SIZE_2D[1] = 0
MODE_SPEC = td.ModeSpec(num_modes=4)
FREQS = [1e14, 2e14]
SOURCES = [
    td.PointDipole(source_time=td.GaussianPulse(freq0=FREQS[0], fwidth=1e14), polarization="Ex"),
    td.ModeSource(
        size=SIZE_2D,
        mode_spec=MODE_SPEC,
        source_time=td.GaussianPulse(freq0=FREQS[1], fwidth=1e14),
        direction="+",
    ),
]
FIELDS = ("Ex", "Ey", "Ez", "Hx", "Hz")
INTERVAL = 2
ORDERS_X = list(range(-1, 2))
ORDERS_Y = list(range(-2, 3))

FS11 = np.linspace(1e14, 2e14, 11)
# unused
# TS = np.linspace(0, 1e-12, 11)
# MODE_INDICES = np.arange(0, 3)
# DIRECTIONS = ["+", "-"]
FS = np.linspace(1e14, 2e14, 5)
TS = np.linspace(0, 1e-12, 4)
MODE_INDICES = np.arange(0, 4)
DIRECTIONS = ["+", "-"]

FIELD_MONITOR = td.FieldMonitor(size=SIZE_3D, fields=FIELDS, name="field", freqs=FREQS)
FIELD_TIME_MONITOR = td.FieldTimeMonitor(
    size=SIZE_3D, fields=FIELDS, name="field_time", interval=INTERVAL
)
FIELD_MONITOR_2D = td.FieldMonitor(size=SIZE_2D, fields=FIELDS, name="field_2d", freqs=FREQS)
FIELD_TIME_MONITOR_2D = td.FieldTimeMonitor(
    size=SIZE_2D, fields=FIELDS, name="field_time_2d", interval=INTERVAL
)
MODE_SOLVE_MONITOR = td.ModeSolverMonitor(
    size=SIZE_2D, name="mode_solver", mode_spec=MODE_SPEC, freqs=FS
)
PERMITTIVITY_MONITOR = td.PermittivityMonitor(size=SIZE_3D, name="permittivity", freqs=FREQS)
MODE_MONITOR = td.ModeMonitor(size=SIZE_2D, name="mode", mode_spec=MODE_SPEC, freqs=FREQS)
FLUX_MONITOR = td.FluxMonitor(size=SIZE_2D, freqs=FREQS, name="flux")
FLUX_TIME_MONITOR = td.FluxTimeMonitor(size=SIZE_2D, interval=INTERVAL, name="flux_time")
DIFFRACTION_MONITOR = td.DiffractionMonitor(
    center=(0, 0, 2),
    size=(td.inf, td.inf, 0),
    freqs=FS,
    name="diffraction",
)

MONITORS = [
    FIELD_MONITOR,
    FIELD_TIME_MONITOR,
    MODE_SOLVE_MONITOR,
    PERMITTIVITY_MONITOR,
    MODE_MONITOR,
    FLUX_MONITOR,
    FLUX_TIME_MONITOR,
    DIFFRACTION_MONITOR,
]

GRID_SPEC = td.GridSpec(wavelength=2.0)
RUN_TIME = 1e-12

SIM_SYM = td.Simulation(
    size=SIZE_3D,
    run_time=RUN_TIME,
    grid_spec=GRID_SPEC,
    symmetry=(1, -1, 1),
    sources=SOURCES,
    monitors=MONITORS,
    structures=STRUCTURES,
    boundary_spec=td.BoundarySpec.all_sides(boundary=td.Periodic()),
)

SIM = td.Simulation(
    size=SIZE_3D,
    run_time=RUN_TIME,
    grid_spec=GRID_SPEC,
    symmetry=(0, 0, 0),
    sources=SOURCES,
    monitors=MONITORS,
    structures=STRUCTURES,
    boundary_spec=td.BoundarySpec.all_sides(boundary=td.Periodic()),
)

""" Generate the data arrays (used in other test files) """


def get_xyz(
    monitor: td.components.monitor.MonitorType, grid_key: str, symmetry: bool
) -> Tuple[List[float], List[float], List[float]]:
    if symmetry:
        grid = SIM_SYM.discretize_monitor(monitor)
        x, y, z = grid[grid_key].to_list
        x = [_x for _x in x if _x >= 0]
        y = [_y for _y in y if _y >= 0]
        z = [_z for _z in z if _z >= 0]
    else:
        grid = SIM.discretize_monitor(monitor)
        x, y, z = grid[grid_key].to_list
    return x, y, z


def make_scalar_field_data_array(grid_key: str, symmetry=True):
    XS, YS, ZS = get_xyz(FIELD_MONITOR, grid_key, symmetry)
    values = (1 + 1j) * np.random.random((len(XS), len(YS), len(ZS), len(FS)))
    return td.ScalarFieldDataArray(values, coords=dict(x=XS, y=YS, z=ZS, f=FS))


def make_scalar_field_time_data_array(grid_key: str, symmetry=True):
    XS, YS, ZS = get_xyz(FIELD_TIME_MONITOR, grid_key, symmetry)
    values = np.random.random((len(XS), len(YS), len(ZS), len(TS)))
    return td.ScalarFieldTimeDataArray(values, coords=dict(x=XS, y=YS, z=ZS, t=TS))


def make_scalar_mode_field_data_array(grid_key: str, symmetry=True):
    XS, YS, ZS = get_xyz(MODE_SOLVE_MONITOR, grid_key, symmetry)
    values = (1 + 0.1j) * np.random.random((len(XS), 1, len(ZS), len(FS), len(MODE_INDICES)))

    return td.ScalarModeFieldDataArray(
        values, coords=dict(x=XS, y=[0.0], z=ZS, f=FS, mode_index=MODE_INDICES)
    )


def make_scalar_mode_field_data_array_smooth(grid_key: str, symmetry=True, rot: float = 0):
    XS, YS, ZS = get_xyz(MODE_SOLVE_MONITOR, grid_key, symmetry)

    values = np.array([1 + 0.1j])[None, :, None, None, None] * np.sin(
        0.5
        * np.pi
        * (MODE_INDICES[None, None, None, None, :] + 1)
        * (1.0 + 3e-15 * (FS[None, None, None, :, None] - FS[0]))
        * (
            np.cos(rot) * np.array(XS)[:, None, None, None, None]
            + np.sin(rot) * np.array(ZS)[None, None, :, None, None]
        )
    )

    return td.ScalarModeFieldDataArray(
        values, coords=dict(x=XS, y=[0.0], z=ZS, f=FS, mode_index=MODE_INDICES)
    )


def make_mode_amps_data_array():
    values = (1 + 1j) * np.random.random((len(DIRECTIONS), len(MODE_INDICES), len(FS)))
    return td.ModeAmpsDataArray(
        values, coords=dict(direction=DIRECTIONS, mode_index=MODE_INDICES, f=FS)
    )


def make_mode_index_data_array():
    values = (1 + 0.1j) * np.random.random((len(FS), len(MODE_INDICES)))
    return td.ModeIndexDataArray(values, coords=dict(f=FS, mode_index=MODE_INDICES))


def make_flux_data_array():
    values = np.random.random(len(FS))
    return td.FluxDataArray(values, coords=dict(f=FS))


def make_flux_time_data_array():
    values = np.random.random(len(TS))
    return td.FluxTimeDataArray(values, coords=dict(t=TS))


def make_diffraction_data_array():
    values = (1 + 1j) * np.random.random((len(ORDERS_X), len(ORDERS_Y), len(FS)))
    return (
        [SIZE_2D[0], SIZE_2D[2]],
        [1.0, 2.0],
        td.DiffractionDataArray(values, coords=dict(orders_x=ORDERS_X, orders_y=ORDERS_Y, f=FS)),
    )


""" Test that they work """


def test_scalar_field_data_array():
    for grid_key in ("Ex", "Ey", "Ez", "Hx", "Hy", "Hz"):
        data = make_scalar_field_data_array(grid_key)
        data = data.interp(f=1.5e14)
        _ = data.isel(y=2)


def test_scalar_field_time_data_array():
    for grid_key in ("Ex", "Ey", "Ez", "Hx", "Hy", "Hz"):
        data = make_scalar_field_time_data_array(grid_key)
        data = data.interp(t=1e-13)
        _ = data.isel(y=2)


def test_scalar_mode_field_data_array():
    for grid_key in ("Ex", "Ey", "Ez", "Hx", "Hy", "Hz"):
        data = make_scalar_mode_field_data_array(grid_key)
        data = data.interp(f=1.5e14)
        data = data.isel(x=2)
        _ = data.sel(mode_index=2)


def test_mode_amps_data_array():
    data = make_mode_amps_data_array()
    data = data.interp(f=1.5e14)
    data = data.isel(direction=0)
    _ = data.sel(mode_index=1)


def test_mode_index_data_array():
    data = make_mode_index_data_array()
    data = data.interp(f=1.5e14)
    _ = data.sel(mode_index=1)


def test_flux_data_array():
    data = make_flux_data_array()
    data = data.interp(f=1.5e14)


def test_flux_time_data_array():
    data = make_flux_time_data_array()
    data = data.interp(t=1e-13)


def test_diffraction_data_array():
    _, _, data = make_diffraction_data_array()
    data = data.interp(f=1.5e14)


def test_ops():
    data1 = make_flux_data_array()
    data2 = make_flux_data_array()
    data1.data = np.ones_like(data1.data)
    data2.data = np.ones_like(data2.data)
    data3 = make_flux_time_data_array()
    assert np.all(data1 == data2), "identical data are not equal"
    data1.data[0] = 1e12
    assert not np.all(data1 == data2), "different data are equal"
    assert not np.all(data1 == data3), "different data are equal"


def test_empty_field_time():
    _ = td.ScalarFieldTimeDataArray(
        np.random.rand(5, 5, 5, 0),
        coords=dict(x=np.arange(5), y=np.arange(5), z=np.arange(5), t=[]),
    )
    _ = td.ScalarFieldTimeDataArray(
        np.random.rand(5, 5, 5, 0),
        coords=dict(x=np.arange(5), y=np.arange(5), z=np.arange(5), t=[]),
    )


def test_abs():
    data = make_mode_amps_data_array()
    _ = data.abs


def test_heat_data_array():
    T = [0, 1e-12, 2e-12]
    _ = td.HeatDataArray((1 + 1j) * np.random.random((3,)), coords=dict(T=T))


def test_charge_data_array():
    n = [0, 1e-12, 2e-12]
    p = [0, 3e-12, 4e-12]
    _ = td.ChargeDataArray((1 + 1j) * np.random.random((3, 3)), coords=dict(n=n, p=p))



"""Tests visualization operations."""
import pytest
import matplotlib.pyplot as plt
import tidy3d as td
from tidy3d.components.viz import Polygon


def test_make_polygon_dict():
    p = Polygon(context={"coordinates": [(1, 0), (0, 1), (0, 0)]})
    p.interiors


@pytest.mark.parametrize("center_z, len_collections", ((0, 1), (0.1, 0)))
def test_0d_plot(center_z, len_collections):
    """Ensure that 0d objects show up in plots."""

    sim = td.Simulation(
        size=(1, 1, 1),
        sources=[
            td.PointDipole(
                center=(0, 0, center_z),
                source_time=td.GaussianPulse(
                    freq0=td.C_0 / 1.0,
                    fwidth=td.C_0 / 5.0,
                ),
                polarization="Ez",
            )
        ],
        run_time=1e-13,
    )

    ax = sim.plot(z=0)

    # if a point is plotted, a single collection will be present, otherwise nothing
    assert len(ax.collections) == len_collections

    plt.close()



"""Tests custom sources and mediums."""
import dill as pickle
from typing import Tuple

import pytest
import numpy as np
import pydantic.v1 as pydantic
import xarray as xr
import tidy3d as td

from ..utils import assert_log_level, log_capture
from tidy3d.components.data.dataset import PermittivityDataset
from tidy3d.components.medium import CustomMedium, CustomPoleResidue, CustomSellmeier
from tidy3d.components.medium import CustomLorentz, CustomDrude, CustomDebye, AbstractCustomMedium
from tidy3d.components.medium import CustomAnisotropicMedium

np.random.seed(4)

Nx, Ny, Nz = 10, 11, 12
X = np.linspace(-1, 1, Nx)
Y = np.linspace(-1, 1, Ny)
Z = np.linspace(-1, 1, Nz)
freqs = [2e14]

ST = td.GaussianPulse(freq0=np.mean(freqs), fwidth=np.mean(freqs) / 5)
SIZE = (2, 0, 2)


def make_scalar_data():
    """Makes a scalar field data array."""
    data = np.random.random((Nx, Ny, Nz, 1)) + 1
    return td.ScalarFieldDataArray(data, coords=dict(x=X, y=Y, z=Z, f=freqs))


def make_scalar_data_multifreqs():
    """Makes a scalar field data array."""
    Nfreq = 2
    freqs_mul = [2e14, 3e14]
    data = np.random.random((Nx, Ny, Nz, Nfreq))
    return td.ScalarFieldDataArray(data, coords=dict(x=X, y=Y, z=Z, f=freqs_mul))


def make_custom_field_source():
    """Make a custom field source."""
    field_components = {}
    for field in "EH":
        for component in "xyz":
            field_components[field + component] = make_scalar_data()
    field_dataset = td.FieldDataset(**field_components)
    return td.CustomFieldSource(size=SIZE, source_time=ST, field_dataset=field_dataset)


def make_custom_current_source():
    """Make a custom field source."""
    field_components = {}
    for field in "EH":
        for component in "xyz":
            field_components[field + component] = make_scalar_data()
    current_dataset = td.FieldDataset(**field_components)
    return td.CustomCurrentSource(size=SIZE, source_time=ST, current_dataset=current_dataset)


# instance, which we use in the parameterized tests
FIELD_SRC = make_custom_field_source()
CURRENT_SRC = make_custom_current_source()


def get_dataset(custom_source_obj) -> Tuple[str, td.FieldDataset]:
    """Get a dict containing dataset depending on type and its key."""
    if isinstance(custom_source_obj, td.CustomFieldSource):
        return "field_dataset", custom_source_obj.field_dataset
    if isinstance(custom_source_obj, td.CustomCurrentSource):
        return "current_dataset", custom_source_obj.current_dataset
    raise ValueError("not supplied a custom current object.")


@pytest.mark.parametrize("source", (FIELD_SRC, CURRENT_SRC))
def test_field_components(source):
    """Get Dictionary of field components and select some data."""
    _, dataset = get_dataset(source)
    for name, field in dataset.field_components.items():
        _ = field.interp(x=0, y=0, z=0).sel(f=freqs[0])


@pytest.mark.parametrize("source", (FIELD_SRC, CURRENT_SRC))
def test_custom_source_simulation(source):
    """Test adding to simulation."""
    _ = td.Simulation(run_time=1e-12, size=(1, 1, 1), sources=(source,))


def test_validator_tangential_field():
    """Test that it errors if no tangential field defined."""
    field_dataset = FIELD_SRC.field_dataset
    field_dataset = field_dataset.copy(update=dict(Ex=None, Ez=None, Hx=None, Hz=None))
    with pytest.raises(pydantic.ValidationError):
        _ = td.CustomFieldSource(size=SIZE, source_time=ST, field_dataset=field_dataset)


def test_validator_non_planar():
    """Test that it errors if the source geometry has a volume."""
    field_dataset = FIELD_SRC.field_dataset
    field_dataset = field_dataset.copy(update=dict(Ex=None, Ez=None, Hx=None, Hz=None))
    with pytest.raises(pydantic.ValidationError):
        _ = td.CustomFieldSource(size=(1, 1, 1), source_time=ST, field_dataset=field_dataset)


@pytest.mark.parametrize("source", (FIELD_SRC, CURRENT_SRC))
def test_validator_freq_out_of_range_src(source):
    """Test that it errors if field_dataset frequency out of range of source_time."""
    key, dataset = get_dataset(source)
    Ex_new = td.ScalarFieldDataArray(dataset.Ex.data, coords=dict(x=X, y=Y, z=Z, f=[0]))
    dataset_fail = dataset.copy(update=dict(Ex=Ex_new))
    with pytest.raises(pydantic.ValidationError):
        _ = source.updated_copy(size=SIZE, source_time=ST, **{key: dataset_fail})


@pytest.mark.parametrize("source", (FIELD_SRC, CURRENT_SRC))
def test_validator_freq_multiple(source):
    """Test that it errors more than 1 frequency given."""
    key, dataset = get_dataset(source)
    new_data = np.concatenate((dataset.Ex.data, dataset.Ex.data), axis=-1)
    Ex_new = td.ScalarFieldDataArray(new_data, coords=dict(x=X, y=Y, z=Z, f=[1, 2]))
    dataset_fail = dataset.copy(update=dict(Ex=Ex_new))
    with pytest.raises(pydantic.ValidationError):
        _ = source.copy(update={key: dataset_fail})


def test_io_hdf5(tmp_path):
    """Saving and loading from hdf5 file."""
    path = str(tmp_path / "custom_source.hdf5")
    FIELD_SRC.to_file(path)
    FIELD_SRC2 = FIELD_SRC.from_file(path)
    assert FIELD_SRC == FIELD_SRC2


def test_io_json(tmp_path, log_capture):
    """to json warns and then from json errors."""
    path = str(tmp_path / "custom_source.json")
    FIELD_SRC.to_file(path)
    assert_log_level(log_capture, "WARNING")
    FIELD_SRC2 = FIELD_SRC.from_file(path)
    assert_log_level(log_capture, "WARNING")
    assert FIELD_SRC2.field_dataset is None


def test_custom_source_pckl(tmp_path):
    path = str(tmp_path / "source.pckl")
    with open(path, "wb") as pickle_file:
        pickle.dump(FIELD_SRC, pickle_file)


def test_io_json_clear_tmp():
    pass


def make_custom_medium(scalar_permittivity_data):
    """Make a custom medium."""
    field_components = {f"eps_{d}{d}": scalar_permittivity_data for d in "xyz"}
    eps_dataset = PermittivityDataset(**field_components)
    return CustomMedium(eps_dataset=eps_dataset)


CUSTOM_MEDIUM = make_custom_medium(make_scalar_data())


def test_medium_components():
    """Get Dictionary of field components and select some data."""
    for name, field in CUSTOM_MEDIUM.eps_dataset.field_components.items():
        _ = field.interp(x=0, y=0, z=0).sel(f=freqs[0])


def test_custom_medium_simulation():
    """Test adding to simulation."""

    struct = td.Structure(
        geometry=td.Box(size=(0.5, 0.5, 0.5)),
        medium=CUSTOM_MEDIUM,
    )

    sim = td.Simulation(
        run_time=1e-12,
        size=(1, 1, 1),
        grid_spec=td.GridSpec.auto(wavelength=1.0),
        structures=(struct,),
    )
    _ = sim.grid


def test_medium_raw():
    """Test from a raw permittivity evaluated at center."""
    eps_raw = make_scalar_data().real
    eps_raw_s = td.SpatialDataArray(eps_raw.squeeze(dim="f", drop=True))

    # lossless
    med = CustomMedium.from_eps_raw(eps_raw)
    meds = CustomMedium.from_eps_raw(eps_raw_s)
    assert med.eps_model(1e14) == meds.eps_model(1e14)

    # lossy
    data = np.random.random((Nx, Ny, Nz, 1)) + 1 + 1e-2 * 1j
    eps_raw = td.ScalarFieldDataArray(data, coords=dict(x=X, y=Y, z=Z, f=freqs))
    eps_raw_s = td.SpatialDataArray(eps_raw.squeeze(dim="f", drop=True))
    med = CustomMedium.from_eps_raw(eps_raw)
    meds = CustomMedium.from_eps_raw(eps_raw_s, freq=freqs[0])
    assert med.eps_model(1e14) == meds.eps_model(1e14)

    # inconsistent freq
    with pytest.raises(td.exceptions.SetupError):
        med = CustomMedium.from_eps_raw(eps_raw, freq=freqs[0] * 1.1)

    # missing freq
    with pytest.raises(td.exceptions.SetupError):
        med = CustomMedium.from_eps_raw(eps_raw_s)


def test_medium_interp():
    """Test if the interp works."""
    coord_interp = td.Coords(**{ax: np.linspace(-2, 2, 20 + ind) for ind, ax in enumerate("xyz")})

    # more than one entries per each axis
    orig_data = make_scalar_data()
    data_fit_nearest = coord_interp.spatial_interp(orig_data, "nearest")
    data_fit_linear = coord_interp.spatial_interp(orig_data, "linear")
    assert np.allclose(data_fit_nearest.shape[:3], [len(f) for f in coord_interp.to_list])
    assert np.allclose(data_fit_linear.shape[:3], [len(f) for f in coord_interp.to_list])
    # maximal or minimal values shouldn't exceed that in the supplied data
    assert max(data_fit_linear.values.ravel()) <= max(orig_data.values.ravel())
    assert min(data_fit_linear.values.ravel()) >= min(orig_data.values.ravel())
    assert max(data_fit_nearest.values.ravel()) <= max(orig_data.values.ravel())
    assert min(data_fit_nearest.values.ravel()) >= min(orig_data.values.ravel())

    # single entry along some axis
    Nx = 1
    X = [1.1]
    data = np.random.random((Nx, Ny, Nz, 1))
    orig_data = td.ScalarFieldDataArray(data, coords=dict(x=X, y=Y, z=Z, f=freqs))
    data_fit_nearest = coord_interp.spatial_interp(orig_data, "nearest")
    data_fit_linear = coord_interp.spatial_interp(orig_data, "linear")
    assert np.allclose(data_fit_nearest.shape[:3], [len(f) for f in coord_interp.to_list])
    assert np.allclose(data_fit_linear.shape[:3], [len(f) for f in coord_interp.to_list])
    # maximal or minimal values shouldn't exceed that in the supplied data
    assert max(data_fit_linear.values.ravel()) <= max(orig_data.values.ravel())
    assert min(data_fit_linear.values.ravel()) >= min(orig_data.values.ravel())
    assert max(data_fit_nearest.values.ravel()) <= max(orig_data.values.ravel())
    assert min(data_fit_nearest.values.ravel()) >= min(orig_data.values.ravel())
    # original data are not modified
    assert not np.allclose(orig_data.shape[:3], [len(f) for f in coord_interp.to_list])


def test_medium_smaller_than_one_positive_sigma():
    """Error when any of eps_inf is lower than 1, or sigma is negative."""
    # single entry along some axis

    # eps_inf < 1
    n_data = 1 + np.random.random((Nx, Ny, Nz, 1))
    n_data[0, 0, 0, 0] = 0.5
    n_dataarray = td.ScalarFieldDataArray(n_data, coords=dict(x=X, y=Y, z=Z, f=freqs))
    with pytest.raises(pydantic.ValidationError):
        _ = CustomMedium.from_nk(n_dataarray)

    # negative sigma
    n_data = 1 + np.random.random((Nx, Ny, Nz, 1))
    k_data = np.random.random((Nx, Ny, Nz, 1))
    k_data[0, 0, 0, 0] = -0.1
    n_dataarray = td.ScalarFieldDataArray(n_data, coords=dict(x=X, y=Y, z=Z, f=freqs))
    k_dataarray = td.ScalarFieldDataArray(k_data, coords=dict(x=X, y=Y, z=Z, f=freqs))
    with pytest.raises(pydantic.ValidationError):
        _ = CustomMedium.from_nk(n_dataarray, k_dataarray)


def test_medium_eps_diagonal_on_grid():
    """Test if ``eps_diagonal_on_grid`` works."""
    coord_interp = td.Coords(**{ax: np.linspace(-1, 1, 20 + ind) for ind, ax in enumerate("xyz")})
    freq_interp = 1e14

    eps_output = CUSTOM_MEDIUM.eps_diagonal_on_grid(freq_interp, coord_interp)

    for i in range(3):
        assert np.allclose(eps_output[i].shape, [len(f) for f in coord_interp.to_list])


def test_medium_nk():
    """Construct custom medium from n (and k) DataArrays."""
    n = make_scalar_data().real
    k = make_scalar_data().real * 0.001
    ns = td.SpatialDataArray(n.squeeze(dim="f", drop=True))
    ks = td.SpatialDataArray(k.squeeze(dim="f", drop=True))

    # lossless
    med = CustomMedium.from_nk(n=n)
    meds = CustomMedium.from_nk(n=ns)
    assert med.eps_model(1e14) == meds.eps_model(1e14)
    # lossy
    med = CustomMedium.from_nk(n=n, k=k)
    meds = CustomMedium.from_nk(n=ns, k=ks, freq=freqs[0])
    assert med.eps_model(1e14) == meds.eps_model(1e14)

    # gain
    with pytest.raises(pydantic.ValidationError):
        med = CustomMedium.from_nk(n=n, k=-k)
    with pytest.raises(pydantic.ValidationError):
        meds = CustomMedium.from_nk(n=ns, k=-ks, freq=freqs[0])
    med = CustomMedium.from_nk(n=n, k=-k, allow_gain=True)
    meds = CustomMedium.from_nk(n=ns, k=-ks, freq=freqs[0], allow_gain=True)
    assert med.eps_model(1e14) == meds.eps_model(1e14)

    # inconsistent freq
    with pytest.raises(td.exceptions.SetupError):
        med = CustomMedium.from_nk(n=n, k=k, freq=freqs[0] * 1.1)

    # missing freq
    with pytest.raises(td.exceptions.SetupError):
        med = CustomMedium.from_nk(n=ns, k=ks)

    # inconsistent data type
    with pytest.raises(td.exceptions.SetupError):
        med = CustomMedium.from_nk(n=ns, k=k)


def test_medium_eps_model():
    """Evaluate the permittivity at a given frequency."""
    med = make_custom_medium(make_scalar_data())
    med.eps_model(frequency=freqs[0])

    # error with multifrequency data
    with pytest.raises(pydantic.ValidationError):
        med = make_custom_medium(make_scalar_data_multifreqs())


def test_nk_diff_coords():
    """Should error if N and K have different coords."""
    n = make_scalar_data().real
    k = make_scalar_data().real
    k.coords["f"] = [3e14]
    with pytest.raises(td.exceptions.SetupError):
        _ = CustomMedium.from_nk(n=n, k=k)


def test_grids():
    """Get Dictionary of field components and select some data."""
    bounds = td.Box(size=(1, 1, 1)).bounds
    for key, grid in CUSTOM_MEDIUM.grids(bounds=bounds).items():
        grid.sizes


def test_n_cfl():
    """CFL number for custom medium"""
    data = np.random.random((Nx, Ny, Nz, 1)) + 2
    ndata = td.ScalarFieldDataArray(data, coords=dict(x=X, y=Y, z=Z, f=freqs))
    med = CustomMedium.from_nk(n=ndata, k=ndata * 0.001)
    assert med.n_cfl >= 2


def verify_custom_medium_methods(mat):
    """Verify that the methods in custom medium is producing expected results."""
    freq = 1.0
    assert isinstance(mat, AbstractCustomMedium)
    assert isinstance(mat.eps_model(freq), np.complex128)
    assert len(mat.eps_diagonal(freq)) == 3
    coord_interp = td.Coords(**{ax: np.linspace(-1, 1, 20 + ind) for ind, ax in enumerate("xyz")})
    eps_grid = mat.eps_diagonal_on_grid(freq, coord_interp)
    for i in range(3):
        assert np.allclose(eps_grid[i].shape, [len(f) for f in coord_interp.to_list])

    # construct sim
    struct = td.Structure(
        geometry=td.Box(size=(0.5, 0.5, 0.5)),
        medium=mat,
    )

    sim = td.Simulation(
        run_time=1e-12,
        size=(1, 1, 1),
        grid_spec=td.GridSpec.auto(wavelength=1.0),
        structures=(struct,),
    )
    _ = sim.grid

    # bkg
    sim = td.Simulation(
        run_time=1e-12,
        size=(1, 1, 1),
        grid_spec=td.GridSpec.auto(wavelength=1.0),
        medium=mat,
    )
    _ = sim.grid


def test_anisotropic_custom_medium():
    """Anisotropic CustomMedium."""

    def make_scalar_data_f():
        """Makes a scalar field data array with random f."""
        data = np.random.random((Nx, Ny, Nz, 1)) + 1
        return td.ScalarFieldDataArray(
            data, coords=dict(x=X, y=Y, z=Z, f=[freqs[0] * np.random.random(1)[0]])
        )

    # same f and different f
    field_components_list = [
        {f"eps_{d}{d}": make_scalar_data() for d in "xyz"},
        {f"eps_{d}{d}": make_scalar_data_f() for d in "xyz"},
    ]
    for field_components in field_components_list:
        eps_dataset = PermittivityDataset(**field_components)
        mat = CustomMedium(eps_dataset=eps_dataset)
        verify_custom_medium_methods(mat)


def test_custom_isotropic_medium():
    """Custom isotropic non-dispersive medium."""
    permittivity = td.SpatialDataArray(
        1 + np.random.random((Nx, Ny, Nz)), coords=dict(x=X, y=Y, z=Z)
    )
    conductivity = td.SpatialDataArray(np.random.random((Nx, Ny, Nz)), coords=dict(x=X, y=Y, z=Z))

    # some terms in permittivity are complex
    with pytest.raises(pydantic.ValidationError):
        epstmp = td.SpatialDataArray(
            1 + np.random.random((Nx, Ny, Nz)) + 0.1j, coords=dict(x=X, y=Y, z=Z)
        )
        mat = CustomMedium(permittivity=epstmp, conductivity=conductivity)

    # some terms in permittivity are < 1
    with pytest.raises(pydantic.ValidationError):
        epstmp = td.SpatialDataArray(np.random.random((Nx, Ny, Nz)), coords=dict(x=X, y=Y, z=Z))
        mat = CustomMedium(permittivity=epstmp, conductivity=conductivity)

    # some terms in conductivity are complex
    with pytest.raises(pydantic.ValidationError):
        sigmatmp = td.SpatialDataArray(
            np.random.random((Nx, Ny, Nz)) + 0.1j, coords=dict(x=X, y=Y, z=Z)
        )
        mat = CustomMedium(permittivity=permittivity, conductivity=sigmatmp)

    # some terms in conductivity are negative
    sigmatmp = td.SpatialDataArray(np.random.random((Nx, Ny, Nz)) - 0.5, coords=dict(x=X, y=Y, z=Z))
    with pytest.raises(pydantic.ValidationError):
        mat = CustomMedium(permittivity=permittivity, conductivity=sigmatmp)
    mat = CustomMedium(permittivity=permittivity, conductivity=sigmatmp, allow_gain=True)
    verify_custom_medium_methods(mat)

    # inconsistent coords
    with pytest.raises(pydantic.ValidationError):
        sigmatmp = td.SpatialDataArray(
            np.random.random((Nx, Ny, Nz)), coords=dict(x=X + 1, y=Y, z=Z)
        )
        mat = CustomMedium(permittivity=permittivity, conductivity=sigmatmp)

    mat = CustomMedium(permittivity=permittivity, conductivity=conductivity)
    verify_custom_medium_methods(mat)

    mat = CustomMedium(permittivity=permittivity)
    verify_custom_medium_methods(mat)


def verify_custom_dispersive_medium_methods(mat):
    """Verify that the methods in custom dispersive medium is producing expected results."""
    verify_custom_medium_methods(mat)
    freq = 1.0
    for i in range(3):
        assert mat.eps_dataarray_freq(freq)[i].shape == (Nx, Ny, Nz)
    np.testing.assert_allclose(mat.eps_model(freq), mat.pole_residue.eps_model(freq))
    coord_interp = td.Coords(**{ax: np.linspace(-1, 1, 20 + ind) for ind, ax in enumerate("xyz")})
    np.testing.assert_allclose(
        mat.eps_diagonal_on_grid(freq, coord_interp),
        mat.pole_residue.eps_diagonal_on_grid(freq, coord_interp),
    )
    for col in range(3):
        for row in range(3):
            np.testing.assert_allclose(
                mat.eps_comp_on_grid(row, col, freq, coord_interp),
                mat.pole_residue.eps_comp_on_grid(row, col, freq, coord_interp),
            )

    # interpolation
    poles_interp = mat.pole_residue.poles_on_grid(coord_interp)
    assert len(poles_interp) == len(mat.pole_residue.poles)
    coord_shape = tuple(len(grid) for grid in coord_interp.to_list)
    for (a, c) in poles_interp:
        assert a.shape == coord_shape
        assert c.shape == coord_shape


def test_custom_pole_residue():
    """Custom pole residue medium."""
    a = td.SpatialDataArray(-np.random.random((Nx, Ny, Nz)), coords=dict(x=X, y=Y, z=Z))
    c = td.SpatialDataArray(np.random.random((Nx, Ny, Nz)) * 1j, coords=dict(x=X, y=Y, z=Z))

    # some terms in eps_inf are negative
    with pytest.raises(pydantic.ValidationError):
        eps_inf = td.SpatialDataArray(
            np.random.random((Nx, Ny, Nz)) - 0.5, coords=dict(x=X, y=Y, z=Z)
        )
        mat = CustomPoleResidue(eps_inf=eps_inf, poles=((a, c),))

    # some terms in eps_inf are complex
    with pytest.raises(pydantic.ValidationError):
        eps_inf = td.SpatialDataArray(
            np.random.random((Nx, Ny, Nz)) + 0.1j, coords=dict(x=X, y=Y, z=Z)
        )
        mat = CustomPoleResidue(eps_inf=eps_inf, poles=((a, c),))

    # inconsistent coords of eps_inf with a,c
    with pytest.raises(pydantic.ValidationError):
        eps_inf = td.SpatialDataArray(
            np.random.random((Nx, Ny, Nz)) + 1, coords=dict(x=X + 1, y=Y, z=Z)
        )
        mat = CustomPoleResidue(eps_inf=eps_inf, poles=((a, c),))

    # break causality
    with pytest.raises(pydantic.ValidationError):
        atmp = td.SpatialDataArray(np.random.random((Nx, Ny, Nz)), coords=dict(x=X, y=Y, z=Z))
        mat = CustomPoleResidue(eps_inf=xr.ones_like(a), poles=((atmp, c),))

    eps_inf = td.SpatialDataArray(np.random.random((Nx, Ny, Nz)) + 1, coords=dict(x=X, y=Y, z=Z))
    mat = CustomPoleResidue(eps_inf=eps_inf, poles=((a, c),))
    verify_custom_dispersive_medium_methods(mat)
    assert mat.n_cfl > 1

    # to custom non-dispersive medium
    # dispersive failure
    with pytest.raises(td.exceptions.ValidationError):
        mat_medium = mat.to_medium()
    # non-dispersive but gain
    a = xr.zeros_like(c)
    mat = CustomPoleResidue(eps_inf=eps_inf, poles=((a, c - 0.1),))
    with pytest.raises(pydantic.ValidationError):
        mat_medium = mat.to_medium()
    mat = CustomPoleResidue(eps_inf=eps_inf, poles=((a, c - 0.1),), allow_gain=True)
    mat_medium = mat.to_medium()
    verify_custom_medium_methods(mat_medium)
    assert mat_medium.n_cfl > 1

    # custom medium to pole residue
    mat = CustomPoleResidue.from_medium(mat_medium)
    verify_custom_dispersive_medium_methods(mat)
    assert mat.n_cfl > 1


def test_custom_sellmeier():
    """Custom Sellmeier medium."""
    b1 = td.SpatialDataArray(np.random.random((Nx, Ny, Nz)), coords=dict(x=X, y=Y, z=Z))
    c1 = td.SpatialDataArray(np.random.random((Nx, Ny, Nz)), coords=dict(x=X, y=Y, z=Z))

    b2 = td.SpatialDataArray(np.random.random((Nx, Ny, Nz)), coords=dict(x=X, y=Y, z=Z))
    c2 = td.SpatialDataArray(np.random.random((Nx, Ny, Nz)), coords=dict(x=X, y=Y, z=Z))

    # complex b
    with pytest.raises(pydantic.ValidationError):
        btmp = td.SpatialDataArray(
            np.random.random((Nx, Ny, Nz)) - 0.5j, coords=dict(x=X, y=Y, z=Z)
        )
        mat = CustomSellmeier(coeffs=((b1, c1), (btmp, c2)))

    # complex c
    with pytest.raises(pydantic.ValidationError):
        ctmp = td.SpatialDataArray(
            np.random.random((Nx, Ny, Nz)) - 0.5j, coords=dict(x=X, y=Y, z=Z)
        )
        mat = CustomSellmeier(coeffs=((b1, c1), (b2, ctmp)))

    # negative c
    with pytest.raises(pydantic.ValidationError):
        ctmp = td.SpatialDataArray(np.random.random((Nx, Ny, Nz)) - 0.5, coords=dict(x=X, y=Y, z=Z))
        mat = CustomSellmeier(coeffs=((b1, c1), (b2, ctmp)))

    # negative b
    btmp = td.SpatialDataArray(np.random.random((Nx, Ny, Nz)) - 0.5, coords=dict(x=X, y=Y, z=Z))
    with pytest.raises(pydantic.ValidationError):
        mat = CustomSellmeier(coeffs=((b1, c1), (btmp, c2)))
    mat = CustomSellmeier(coeffs=((b1, c1), (btmp, c2)), allow_gain=True)
    assert mat.pole_residue.allow_gain

    # inconsistent coord
    with pytest.raises(pydantic.ValidationError):
        btmp = td.SpatialDataArray(np.random.random((Nx, Ny, Nz)), coords=dict(x=X + 1, y=Y, z=Z))
        mat = CustomSellmeier(coeffs=((b1, c2), (btmp, c2)))

    mat = CustomSellmeier(coeffs=((b1, c1), (b2, c2)))
    verify_custom_dispersive_medium_methods(mat)
    assert mat.n_cfl == 1

    # from dispersion
    n = td.SpatialDataArray(2 + np.random.random((Nx, Ny, Nz)), coords=dict(x=X, y=Y, z=Z))
    dn_dwvl = td.SpatialDataArray(-np.random.random((Nx, Ny, Nz)), coords=dict(x=X, y=Y, z=Z))
    mat = CustomSellmeier.from_dispersion(n=n, dn_dwvl=dn_dwvl, freq=2, interp_method="linear")
    verify_custom_dispersive_medium_methods(mat)
    assert mat.n_cfl == 1


def test_custom_lorentz():
    """Custom Lorentz medium."""
    eps_inf = td.SpatialDataArray(np.random.random((Nx, Ny, Nz)) + 1, coords=dict(x=X, y=Y, z=Z))

    de1 = td.SpatialDataArray(np.random.random((Nx, Ny, Nz)), coords=dict(x=X, y=Y, z=Z))
    f1 = td.SpatialDataArray(1 + np.random.random((Nx, Ny, Nz)), coords=dict(x=X, y=Y, z=Z))
    delta1 = td.SpatialDataArray(np.random.random((Nx, Ny, Nz)), coords=dict(x=X, y=Y, z=Z))

    de2 = td.SpatialDataArray(np.random.random((Nx, Ny, Nz)), coords=dict(x=X, y=Y, z=Z))
    f2 = td.SpatialDataArray(1 + np.random.random((Nx, Ny, Nz)), coords=dict(x=X, y=Y, z=Z))
    delta2 = td.SpatialDataArray(np.random.random((Nx, Ny, Nz)), coords=dict(x=X, y=Y, z=Z))

    # complex de
    with pytest.raises(pydantic.ValidationError):
        detmp = td.SpatialDataArray(
            np.random.random((Nx, Ny, Nz)) - 0.5j, coords=dict(x=X, y=Y, z=Z)
        )
        mat = CustomLorentz(eps_inf=eps_inf, coeffs=((de1, f1, delta1), (detmp, f2, delta2)))

    # mixed delta > f and delta < f over spatial points
    with pytest.raises(pydantic.ValidationError):
        deltatmp = td.SpatialDataArray(
            1 + np.random.random((Nx, Ny, Nz)), coords=dict(x=X, y=Y, z=Z)
        )
        mat = CustomLorentz(eps_inf=eps_inf, coeffs=((de1, f1, delta1), (de2, f2, deltatmp)))

    # inconsistent coords
    with pytest.raises(pydantic.ValidationError):
        ftmp = td.SpatialDataArray(
            1 + np.random.random((Nx, Ny, Nz)), coords=dict(x=X + 1, y=Y, z=Z)
        )
        mat = CustomLorentz(eps_inf=eps_inf, coeffs=((de1, f1, delta1), (de2, ftmp, delta2)))

    # break causality with negative delta
    with pytest.raises(pydantic.ValidationError):
        deltatmp = td.SpatialDataArray(
            np.random.random((Nx, Ny, Nz)) - 0.5, coords=dict(x=X, y=Y, z=Z)
        )
        mat = CustomLorentz(eps_inf=eps_inf, coeffs=((de1, f1, delta1), (de2, f2, deltatmp)))

    # gain medium with negative delta epsilon
    with pytest.raises(pydantic.ValidationError):
        detmp = td.SpatialDataArray(
            np.random.random((Nx, Ny, Nz)) - 0.5, coords=dict(x=X, y=Y, z=Z)
        )
        mat = CustomLorentz(eps_inf=eps_inf, coeffs=((de1, f1, delta1), (detmp, f2, delta2)))
    mat = CustomLorentz(
        eps_inf=eps_inf, coeffs=((de1, f1, delta1), (detmp, f2, delta2)), allow_gain=True
    )
    verify_custom_dispersive_medium_methods(mat)
    assert mat.n_cfl > 1

    mat = CustomLorentz(
        eps_inf=eps_inf, coeffs=((de1, f1, delta1), (de2, f2, delta2)), subpixel=True
    )
    verify_custom_dispersive_medium_methods(mat)
    assert mat.n_cfl > 1
    assert mat.pole_residue.subpixel


def test_custom_drude():
    """Custom Drude medium."""
    eps_inf = td.SpatialDataArray(np.random.random((Nx, Ny, Nz)) + 1, coords=dict(x=X, y=Y, z=Z))

    f1 = td.SpatialDataArray(1 + np.random.random((Nx, Ny, Nz)), coords=dict(x=X, y=Y, z=Z))
    delta1 = td.SpatialDataArray(np.random.random((Nx, Ny, Nz)), coords=dict(x=X, y=Y, z=Z))

    f2 = td.SpatialDataArray(1 + np.random.random((Nx, Ny, Nz)), coords=dict(x=X, y=Y, z=Z))
    delta2 = td.SpatialDataArray(np.random.random((Nx, Ny, Nz)), coords=dict(x=X, y=Y, z=Z))

    # complex delta
    with pytest.raises(pydantic.ValidationError):
        deltatmp = td.SpatialDataArray(
            np.random.random((Nx, Ny, Nz)) - 0.5j, coords=dict(x=X, y=Y, z=Z)
        )
        mat = CustomDrude(eps_inf=eps_inf, coeffs=((f1, delta1), (f2, deltatmp)))

    # negative delta
    with pytest.raises(pydantic.ValidationError):
        deltatmp = td.SpatialDataArray(
            np.random.random((Nx, Ny, Nz)) - 0.5, coords=dict(x=X, y=Y, z=Z)
        )
        mat = CustomDrude(eps_inf=eps_inf, coeffs=((f1, delta1), (f2, deltatmp)))

    # inconsistent coords
    with pytest.raises(pydantic.ValidationError):
        ftmp = td.SpatialDataArray(
            1 + np.random.random((Nx, Ny, Nz)), coords=dict(x=X + 1, y=Y, z=Z)
        )
        mat = CustomDrude(eps_inf=eps_inf, coeffs=((f1, delta1), (ftmp, delta2)))

    mat = CustomDrude(eps_inf=eps_inf, coeffs=((f1, delta1), (f2, delta2)))
    verify_custom_dispersive_medium_methods(mat)
    assert mat.n_cfl > 1


def test_custom_debye():
    """Custom Debye medium."""
    eps_inf = td.SpatialDataArray(np.random.random((Nx, Ny, Nz)) + 1, coords=dict(x=X, y=Y, z=Z))

    eps1 = td.SpatialDataArray(np.random.random((Nx, Ny, Nz)), coords=dict(x=X, y=Y, z=Z))
    tau1 = td.SpatialDataArray(np.random.random((Nx, Ny, Nz)), coords=dict(x=X, y=Y, z=Z))

    eps2 = td.SpatialDataArray(np.random.random((Nx, Ny, Nz)), coords=dict(x=X, y=Y, z=Z))
    tau2 = td.SpatialDataArray(np.random.random((Nx, Ny, Nz)), coords=dict(x=X, y=Y, z=Z))

    # complex eps
    with pytest.raises(pydantic.ValidationError):
        epstmp = td.SpatialDataArray(
            np.random.random((Nx, Ny, Nz)) - 0.5j, coords=dict(x=X, y=Y, z=Z)
        )
        mat = CustomDebye(eps_inf=eps_inf, coeffs=((eps1, tau1), (epstmp, tau2)))

    # complex tau
    with pytest.raises(pydantic.ValidationError):
        tautmp = td.SpatialDataArray(
            np.random.random((Nx, Ny, Nz)) - 0.5j, coords=dict(x=X, y=Y, z=Z)
        )
        mat = CustomDebye(eps_inf=eps_inf, coeffs=((eps1, tau1), (eps2, tautmp)))

    # negative tau
    with pytest.raises(pydantic.ValidationError):
        tautmp = td.SpatialDataArray(
            np.random.random((Nx, Ny, Nz)) - 0.5, coords=dict(x=X, y=Y, z=Z)
        )
        mat = CustomDebye(eps_inf=eps_inf, coeffs=((eps1, tau1), (eps2, tautmp)))

    # inconsistent coords
    with pytest.raises(pydantic.ValidationError):
        epstmp = td.SpatialDataArray(np.random.random((Nx, Ny, Nz)), coords=dict(x=X + 1, y=Y, z=Z))
        mat = CustomDebye(eps_inf=eps_inf, coeffs=((eps1, tau1), (epstmp, tau2)))

    # negative delta epsilon
    with pytest.raises(pydantic.ValidationError):
        epstmp = td.SpatialDataArray(
            np.random.random((Nx, Ny, Nz)) - 0.5, coords=dict(x=X, y=Y, z=Z)
        )
        mat = CustomDebye(eps_inf=eps_inf, coeffs=((eps1, tau1), (epstmp, tau2)))
    mat = CustomDebye(eps_inf=eps_inf, coeffs=((eps1, tau1), (epstmp, tau2)), allow_gain=True)
    verify_custom_dispersive_medium_methods(mat)
    assert mat.n_cfl > 1

    mat = CustomDebye(eps_inf=eps_inf, coeffs=((eps1, tau1), (eps2, tau2)))
    verify_custom_dispersive_medium_methods(mat)
    assert mat.n_cfl > 1


def test_custom_anisotropic_medium(log_capture):
    """Custom anisotropic medium."""

    # xx
    permittivity = td.SpatialDataArray(
        1 + np.random.random((Nx, Ny, Nz)), coords=dict(x=X, y=Y, z=Z)
    )
    conductivity = td.SpatialDataArray(np.random.random((Nx, Ny, Nz)), coords=dict(x=X, y=Y, z=Z))
    mat_xx = CustomMedium(permittivity=permittivity, conductivity=conductivity)

    # yy
    eps_inf = td.SpatialDataArray(np.random.random((Nx, Ny, Nz)) + 1, coords=dict(x=X, y=Y, z=Z))
    eps1 = td.SpatialDataArray(np.random.random((Nx, Ny, Nz)), coords=dict(x=X, y=Y, z=Z))
    tau1 = td.SpatialDataArray(np.random.random((Nx, Ny, Nz)), coords=dict(x=X, y=Y, z=Z))
    eps2 = td.SpatialDataArray(np.random.random((Nx, Ny, Nz)), coords=dict(x=X, y=Y, z=Z))
    tau2 = td.SpatialDataArray(np.random.random((Nx, Ny, Nz)), coords=dict(x=X, y=Y, z=Z))
    mat_yy = CustomDebye(eps_inf=eps_inf, coeffs=((eps1, tau1), (eps2, tau2)))

    # zz
    eps_inf = td.SpatialDataArray(np.random.random((Nx, Ny, Nz)) + 1, coords=dict(x=X, y=Y, z=Z))
    f1 = td.SpatialDataArray(1 + np.random.random((Nx, Ny, Nz)), coords=dict(x=X, y=Y, z=Z))
    delta1 = td.SpatialDataArray(np.random.random((Nx, Ny, Nz)), coords=dict(x=X, y=Y, z=Z))
    f2 = td.SpatialDataArray(1 + np.random.random((Nx, Ny, Nz)), coords=dict(x=X, y=Y, z=Z))
    delta2 = td.SpatialDataArray(np.random.random((Nx, Ny, Nz)), coords=dict(x=X, y=Y, z=Z))
    mat_zz = CustomDrude(eps_inf=eps_inf, coeffs=((f1, delta1), (f2, delta2)))

    # anisotropic
    mat = CustomAnisotropicMedium(xx=mat_xx, yy=mat_yy, zz=mat_zz)
    verify_custom_medium_methods(mat)

    mat = CustomAnisotropicMedium(xx=mat_xx, yy=mat_yy, zz=mat_zz, subpixel=True)
    assert_log_level(log_capture, "WARNING")

    ## interpolation method verification for "xx" component
    # 1) xx-component is using `interp_method = nearest`, and mat using `None`;
    # so that xx-component is using "nearest"
    freq = 2e14
    dist_coeff = 0.6
    coord_test = td.Coords(x=[X[0] * dist_coeff + X[1] * (1 - dist_coeff)], y=Y[0], z=Z[0])
    eps_nearest = mat.eps_sigma_to_eps_complex(
        permittivity.sel(x=X[0], y=Y[0], z=Z[0]), conductivity.sel(x=X[0], y=Y[0], z=Z[0]), freq
    )
    eps_interp = mat.eps_comp_on_grid(0, 0, freq, coord_test)[0, 0, 0]
    assert eps_interp == eps_nearest.data

    # 2) xx-component is using `interp_method = nearest`, and mat using `nearest`;
    # so that xx-component is still using "nearest"
    mat = CustomAnisotropicMedium(xx=mat_xx, yy=mat_yy, zz=mat_zz, interp_method="nearest")
    eps_interp = mat.eps_comp_on_grid(0, 0, freq, coord_test)[0, 0, 0]
    assert eps_interp == eps_nearest.data

    # 3) xx-component is using `interp_method = nearest`, and mat using `linear`;
    # so that xx-component is using "linear" (overridden by the one in mat)
    mat = CustomAnisotropicMedium(xx=mat_xx, yy=mat_yy, zz=mat_zz, interp_method="linear")
    eps_second_nearest = mat.eps_sigma_to_eps_complex(
        permittivity.sel(x=X[1], y=Y[0], z=Z[0]), conductivity.sel(x=X[1], y=Y[0], z=Z[0]), freq
    )
    eps_manual_interp = eps_nearest * dist_coeff + eps_second_nearest * (1 - dist_coeff)
    eps_interp = mat.eps_comp_on_grid(0, 0, freq, coord_test)[0, 0, 0]
    assert np.isclose(eps_interp, eps_manual_interp.data)

    # 4) xx-component is using `interp_method = linear`, and mat using `None`;
    # so that xx-component is using "linear"
    mat_xx = CustomMedium(
        permittivity=permittivity, conductivity=conductivity, interp_method="linear"
    )
    mat = CustomAnisotropicMedium(xx=mat_xx, yy=mat_yy, zz=mat_zz)
    eps_interp = mat.eps_comp_on_grid(0, 0, freq, coord_test)[0, 0, 0]
    assert np.isclose(eps_interp, eps_manual_interp.data)

    # 5) xx-component is using `interp_method = linear`, and mat using `linear`;
    # so that xx-component is still using "linear"
    mat = CustomAnisotropicMedium(xx=mat_xx, yy=mat_yy, zz=mat_zz, interp_method="linear")
    eps_interp = mat.eps_comp_on_grid(0, 0, freq, coord_test)[0, 0, 0]
    assert np.isclose(eps_interp, eps_manual_interp.data)

    # 6) xx-component is using `interp_method = linear`, and mat using `nearest`;
    # so that xx-component is using "nearest" (overridden by the one in mat)
    mat = CustomAnisotropicMedium(xx=mat_xx, yy=mat_yy, zz=mat_zz, interp_method="nearest")
    eps_interp = mat.eps_comp_on_grid(0, 0, freq, coord_test)[0, 0, 0]
    assert eps_interp == eps_nearest.data

    # CustomMedium is anisotropic
    field_components = {f"eps_{d}{d}": make_scalar_data() for d in "xyz"}
    eps_dataset = PermittivityDataset(**field_components)
    mat_tmp = CustomMedium(eps_dataset=eps_dataset)
    with pytest.raises(pydantic.ValidationError):
        mat = CustomAnisotropicMedium(xx=mat_tmp, yy=mat_yy, zz=mat_zz)
    with pytest.raises(pydantic.ValidationError):
        mat = CustomAnisotropicMedium(xx=mat_xx, yy=mat_tmp, zz=mat_zz)
    with pytest.raises(pydantic.ValidationError):
        mat = CustomAnisotropicMedium(xx=mat_xx, yy=mat_yy, zz=mat_tmp)


def test_io_dispersive(tmp_path):
    Mx_custom = 1.0
    My_custom = 2.1
    Nx = 10
    Ny = 11

    x_custom = np.linspace(-Mx_custom / 2, Mx_custom / 2, Nx)
    y_custom = np.linspace(-My_custom / 2, My_custom / 2, Ny)
    z_custom = [0]

    delep_data = np.ones([len(x_custom), len(y_custom), len(z_custom)])
    delep_dataset = td.SpatialDataArray(
        delep_data, coords={"x": x_custom, "y": y_custom, "z": z_custom}
    )
    gamma_dataset = xr.zeros_like(delep_dataset)
    f0_dataset = td.SpatialDataArray(
        np.ones_like(delep_data) * 3e14, coords={"x": x_custom, "y": y_custom, "z": z_custom}
    )
    eps_inf_dataset = xr.ones_like(delep_dataset)
    mat_custom = td.CustomLorentz(
        eps_inf=eps_inf_dataset, coeffs=((delep_dataset, f0_dataset, gamma_dataset),)
    )

    struct = td.Structure(
        geometry=td.Box(size=(0.5, 0.5, 0.5)),
        medium=mat_custom,
    )

    sim = td.Simulation(
        run_time=1e-12,
        size=(1, 1, 1),
        grid_spec=td.GridSpec.auto(wavelength=1.0),
        structures=(struct,),
    )

    filename = str(tmp_path / "sim.hdf5")
    sim.to_file(filename)
    sim_load = td.Simulation.from_file(filename)

    assert sim_load == sim



"""Test running tidy3d as command line application."""

import tidy3d as td
import pytest
from tidy3d.__main__ import main


def save_sim_to_path(path: str) -> None:
    sim = td.Simulation(size=(1, 1, 1), grid_spec=td.GridSpec.auto(wavelength=1.0), run_time=1e-12)
    sim.to_file(path)


@pytest.mark.parametrize("extension", (".json", ".yaml"))
def test_main(extension, tmp_path):
    path = str((tmp_path / "sim").with_suffix(extension))
    save_sim_to_path(path)
    main([path, "--test_only"])



"""Test the logging."""

import pytest
import json

import pydantic.v1 as pd
import numpy as np
import tidy3d as td
from tidy3d.exceptions import Tidy3dError
from tidy3d.log import DEFAULT_LEVEL, _get_level_int, set_logging_level


def test_log():
    td.log.debug("debug test")
    td.log.info("info test")
    td.log.warning("warning test")
    td.log.error("error test")
    td.log.critical("critical test")
    td.log.log(0, "zero test")


def test_log_config(tmp_path):
    td.config.logging_level = "DEBUG"
    td.set_logging_file(str(tmp_path / "test.log"))
    assert len(td.log.handlers) == 2
    assert td.log.handlers["console"].level == _get_level_int("DEBUG")
    assert td.log.handlers["file"].level == _get_level_int(DEFAULT_LEVEL)
    del td.log.handlers["file"]


def test_log_level_not_found():
    with pytest.raises(ValueError):
        set_logging_level("NOT_A_LEVEL")


def test_set_logging_level_deprecated():
    with pytest.raises(DeprecationWarning):
        td.set_logging_level("WARNING")


def test_exception_message():
    MESSAGE = "message"
    e = Tidy3dError(MESSAGE)
    assert str(e) == MESSAGE


def test_logging_upper():
    """Make sure we get an error if lowercase."""
    td.config.logging_level = "WARNING"
    with pytest.raises(ValueError):
        td.config.logging_level = "warning"


def test_logging_unrecognized():
    """If unrecognized option, raise validation errorr."""
    with pytest.raises(pd.ValidationError):
        td.config.logging_level = "blah"


def test_logging_warning_capture():
    # create sim with warnings
    domain_size = 12

    wavelength = 1
    f0 = td.C_0 / wavelength
    fwidth = f0 / 10.0
    source_time = td.GaussianPulse(freq0=f0, fwidth=fwidth)
    freqs = np.linspace(f0 - fwidth, f0 + fwidth, 11)

    # 1 warning: too long run_time
    run_time = 10000 / fwidth

    # 1 warning: frequency outside of source frequency range
    mode_mnt = td.ModeMonitor(
        center=(0, 0, 0),
        size=(domain_size, 0, domain_size),
        freqs=list(freqs) + [0.1],
        mode_spec=td.ModeSpec(num_modes=3),
        name="mode",
    )

    # 1 warning: too high num_freqs
    mode_source = td.ModeSource(
        size=(domain_size, 0, domain_size),
        source_time=source_time,
        mode_spec=td.ModeSpec(num_modes=2, precision="single"),
        mode_index=1,
        num_freqs=50,
        direction="-",
    )

    # 1 warning: ignoring "normal_dir"
    monitor_flux = td.FluxMonitor(
        center=(0, 0, 0),
        size=(8, 8, 8),
        freqs=list(freqs),
        name="flux",
        normal_dir="+",
    )

    # 1 warning: large monitor size
    monitor_time = td.FieldTimeMonitor(
        center=(0, 0, 0),
        size=(2, 2, 2),
        stop=1 / fwidth,
        name="time",
    )

    # 1 warning: too big proj distance
    proj_mnt = td.FieldProjectionCartesianMonitor(
        center=(0, 0, 0),
        size=(2, 2, 2),
        freqs=[250e12, 300e12],
        name="n2f_monitor",
        custom_origin=(1, 2, 3),
        x=[-1, 0, 1],
        y=[-2, -1, 0, 1, 2],
        proj_axis=2,
        proj_distance=1e10,
        far_field_approx=False,
    )

    # 2 warnings * 4 sources = 8 total: too close to each PML
    # 1 warning * 3 DFT monitors = 3 total: medium frequency range does not cover monitors freqs
    box = td.Structure(
        geometry=td.Box(center=(0, 0, 0), size=(11.5, 11.5, 11.5)),
        medium=td.Medium(permittivity=2, frequency_range=[0.5, 1]),
    )

    # 2 warnings: inside pml
    box_in_pml = td.Structure(
        geometry=td.Box(center=(0, 0, 0), size=(domain_size * 1.001, 5, 5)),
        medium=td.Medium(permittivity=10),
    )

    # 2 warnings: exactly on sim edge
    box_on_boundary = td.Structure(
        geometry=td.Box(center=(0, 0, 0), size=(domain_size, 5, 5)),
        medium=td.Medium(permittivity=20),
    )

    # 1 warning: outside of domain
    box_outside = td.Structure(
        geometry=td.Box(center=(50, 0, 0), size=(domain_size, 5, 5)),
        medium=td.Medium(permittivity=6),
    )

    # 1 warning: too high "num_freqs"
    # 1 warning: glancing angle
    gaussian_beam = td.GaussianBeam(
        center=(4, 0, 0),
        size=(0, 2, 1),
        waist_radius=2.0,
        waist_distance=1,
        source_time=source_time,
        direction="+",
        num_freqs=30,
        angle_theta=np.pi / 2.1,
    )

    plane_wave = td.PlaneWave(
        center=(4, 0, 0),
        size=(0, 1, 2),
        source_time=source_time,
        direction="+",
    )

    # 2 warnings: non-uniform grid along y and z
    tfsf = td.TFSF(
        size=(10, 15, 15),
        source_time=source_time,
        direction="-",
        injection_axis=0,
    )

    # 1 warning: bloch boundary is inconsistent with plane_wave
    bspec = td.BoundarySpec(
        x=td.Boundary.pml(), y=td.Boundary.periodic(), z=td.Boundary.bloch(bloch_vec=0.2)
    )

    # 1 warning * 1 structures (perm=20) * 4 sources = 20 total: large grid step along x
    gspec = td.GridSpec(
        grid_x=td.UniformGrid(dl=0.05),
        grid_y=td.AutoGrid(min_steps_per_wvl=15),
        grid_z=td.AutoGrid(min_steps_per_wvl=15),
        override_structures=[
            td.Structure(geometry=td.Box(size=(3, 2, 1)), medium=td.Medium(permittivity=4))
        ],
    )

    sim = td.Simulation(
        size=[domain_size, 20, 20],
        sources=[gaussian_beam, mode_source, plane_wave, tfsf],
        structures=[box, box_in_pml, box_on_boundary, box_outside],
        # monitors=[monitor_flux, mode_mnt, monitor_time, proj_mnt],
        monitors=[monitor_flux, mode_mnt, proj_mnt],
        run_time=run_time,
        boundary_spec=bspec,
        grid_spec=gspec,
    )

    # parse the entire simulation at once to capture warnings hierarchically
    sim_dict = sim.dict()

    # re-add projection monitors because it has been overwritten in validators (far_field_approx=False -> True)
    monitors = list(sim_dict["monitors"])
    monitors[2] = proj_mnt.dict()

    sim_dict["monitors"] = monitors

    td.log.set_capture(True)
    sim = td.Simulation.parse_obj(sim_dict)
    print(sim.monitors_data_size)
    sim.validate_pre_upload()
    warning_list = td.log.captured_warnings()
    print(json.dumps(warning_list, indent=4))
    assert len(warning_list) == 30
    td.log.set_capture(False)

    # check that capture doesn't change validation errors

    # validation error during parse_obj()
    sim_dict_no_source = sim.dict()
    sim_dict_no_source.update({"sources": []})

    # validation error during validate_pre_upload()
    sim_dict_large_mnt = sim.dict()
    sim_dict_large_mnt.update({"monitors": [monitor_time.updated_copy(size=(10, 10, 10))]})

    # for sim_dict in [sim_dict_no_source, sim_dict_large_mnt]:
    for sim_dict in [sim_dict_no_source]:

        try:
            sim = td.Simulation.parse_obj(sim_dict)
            sim.validate_pre_upload()
        except pd.ValidationError as e:
            error_without = e.errors()
        except Exception as e:
            error_without = str(e)

        td.log.set_capture(True)
        try:
            sim = td.Simulation.parse_obj(sim_dict)
            sim.validate_pre_upload()
        except pd.ValidationError as e:
            error_with = e.errors()
        except Exception as e:
            error_with = str(e)
        td.log.set_capture(False)

        print(error_without)
        print(error_with)

        assert error_without == error_with


def test_log_suppression():
    with td.log as suppressed_log:
        assert td.log._counts is not None
        for i in range(4):
            suppressed_log.warning("Warning message")
        assert td.log._counts[30] == 3

    td.config.log_suppression = False
    with td.log as suppressed_log:
        assert td.log._counts is None
        for i in range(4):
            suppressed_log.warning("Warning message")
        assert td.log._counts is None

    td.config.log_suppression = True
